#!/usr/bin/env python3
# cardinal_pythonlib/openxml/grep_in_openxml.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Performs a grep (global-regular-expression-print) search of files in OpenXML
format, which is to say inside ZIP files. See the command-line help for
details.**

Version history:

- Written 28 Sep 2017.

Notes:

- use the ``vbindiff`` tool to show *how* two binary files differ.

"""

from argparse import ArgumentParser, RawDescriptionHelpFormatter
import logging
import multiprocessing
import os
import re
from sys import argv, getdefaultencoding, stdin
from typing import Pattern
from zipfile import BadZipFile, ZipFile
import zlib

from cardinal_pythonlib.logs import (
    BraceStyleAdapter,
    main_only_quicksetup_rootlogger,
)
from cardinal_pythonlib.fileops import gen_filenames

log = BraceStyleAdapter(logging.getLogger(__name__))


def report_hit_filename(zipfilename: str, contentsfilename: str,
                        show_inner_file: bool) -> None:
    """
    For "hits": prints either the ``.zip`` filename, or the ``.zip`` filename
    and the inner filename.

    Args:
        zipfilename: filename of the ``.zip`` file
        contentsfilename: filename of the inner file
        show_inner_file: if ``True``, show both; if ``False``, show just the
            ``.zip`` filename

    Returns:

    """
    if show_inner_file:
        print(f"{zipfilename} [{contentsfilename}]")
    else:
        print(zipfilename)


def report_miss_filename(zipfilename: str) -> None:
    """
    For "misses": prints the zip filename.
    """
    print(zipfilename)


def report_line(zipfilename: str, contentsfilename: str, line: str,
                show_inner_file: bool) -> None:
    """
    Prints a line from a file, with the ``.zip`` filename and optionally also
    the inner filename.

    Args:
        zipfilename: filename of the ``.zip`` file
        contentsfilename: filename of the inner file
        line: the line from the inner file
        show_inner_file: if ``True``, show both filenames; if ``False``, show
            just the ``.zip`` filename
    """
    if show_inner_file:
        print(f"{zipfilename} [{contentsfilename}]: {line}")
    else:
        print(f"{zipfilename}: {line}")


def parse_zip(zipfilename: str,
              regex: Pattern,
              invert_match: bool,
              files_with_matches: bool,
              files_without_match: bool,
              grep_inner_file_name: bool,
              show_inner_file: bool) -> None:
    """
    Implement a "grep within an OpenXML file" for a single OpenXML file, which
    is by definition a ``.zip`` file.

    Args:
        zipfilename: name of the OpenXML (zip) file
        regex: regular expression to match
        invert_match: find files that do NOT match, instead of ones that do?
        files_with_matches: show filenames of files with a match?
        files_without_match: show filenames of files with no match?
        grep_inner_file_name: search the names of "inner" files, rather than
            their contents?
        show_inner_file: show the names of the "inner" files, not just the
            "outer" (OpenXML) file?
    """
    assert not (files_without_match and files_with_matches)
    report_lines = (not files_without_match) and (not files_with_matches)
    report_hit_lines = report_lines and not invert_match
    report_miss_lines = report_lines and invert_match
    log.debug("Checking ZIP: " + zipfilename)
    found_in_zip = False
    try:
        with ZipFile(zipfilename, 'r') as zf:
            for contentsfilename in zf.namelist():
                log.debug("... checking file: " + contentsfilename)
                if grep_inner_file_name:
                    found_in_filename = bool(regex.search(contentsfilename))
                    found_in_zip = found_in_zip or found_in_filename
                    if files_with_matches and found_in_zip:
                        report_hit_filename(zipfilename, contentsfilename,
                                            show_inner_file)
                        return
                    if ((report_hit_lines and found_in_filename) or
                            (report_miss_lines and not found_in_filename)):
                        report_line(zipfilename, contentsfilename,
                                    contentsfilename, show_inner_file)
                else:
                    try:
                        with zf.open(contentsfilename, 'r') as file:
                            try:
                                for line in file.readlines():
                                    # log.debug("line: {!r}", line)
                                    found_in_line = bool(regex.search(line))
                                    found_in_zip = found_in_zip or found_in_line
                                    if files_with_matches and found_in_zip:
                                        report_hit_filename(zipfilename,
                                                            contentsfilename,
                                                            show_inner_file)
                                        return
                                    if ((report_hit_lines and found_in_line) or
                                            (report_miss_lines and
                                             not found_in_line)):
                                        report_line(zipfilename,
                                                    contentsfilename,
                                                    line, show_inner_file)
                            except EOFError:
                                pass
                    except RuntimeError as e:
                        log.warning(
                            "RuntimeError whilst processing {} [{}]: probably "
                            "encrypted contents; error was {!r}",
                            zipfilename, contentsfilename, e)
    except (zlib.error, BadZipFile) as e:
        log.debug("Invalid zip: {}; error was {!r}", zipfilename, e)
    if files_without_match and not found_in_zip:
        report_miss_filename(zipfilename)


def main() -> None:
    """
    Command-line handler for the ``grep_in_openxml`` tool.
    Use the ``--help`` option for help.
    """
    exe_name = os.path.basename(argv[0]) or "grep_in_openxml"
    parser = ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description=f"""
Performs a grep (global-regular-expression-print) search of files in OpenXML
format, which is to say inside ZIP files.

Note that you can chain; for example, to search for OpenXML files containing
both "armadillo" and "bonobo", you can do:

    {exe_name} -l armadillo *.pptx | grep_in_openxml -x -l bonobo

To find files in a tree, you can use the "find" tool. For example, to find all
".docx" files in a directory (or its subdirectories) that contain the phrase
"armadillo country", you could use:

    find <STARTDIR> -type f -name "*.docx" -exec {exe_name} -l "armadillo country" {{}} \;

Or you could use this tool directly, specifying a directory and "--recursive",
as in

    {exe_name} -l --recursive "armadillo country" <STARTDIR>

but that won't restrict to ".docx" files.
"""  # noqa
    )
    parser.add_argument(
        "pattern",
        help="Regular expression pattern to apply."
    )
    parser.add_argument(
        "filename", nargs="*",
        help="File(s) to check. You can also specify directores if you use "
             "--recursive"
    )
    parser.add_argument(
        "--filenames_from_stdin", "-x", action="store_true",
        help="Take filenames from stdin instead, one line per filename "
             "(useful for chained grep)."
    )
    parser.add_argument(
        "--recursive", action="store_true",
        help="Allow search to descend recursively into any directories "
             "encountered."
    )
    # Flag abbreviations to match grep:
    parser.add_argument(
        "--ignore_case", "-i", action="store_true",
        help="Ignore case"
    )
    parser.add_argument(
        "--invert_match", "-v", action="store_true",
        help="Invert match"
    )
    parser.add_argument(
        "--files_with_matches", "-l", action="store_true",
        help="Show filenames of files with matches"
    )
    parser.add_argument(
        "--files_without_match", "-L", action="store_true",
        help="Show filenames of files with no match"
    )
    parser.add_argument(
        "--grep_inner_file_name", action="store_true",
        help="Search the NAMES of the inner files, not their contents."
    )
    parser.add_argument(
        "--show_inner_file", action="store_true",
        help="For hits, show the filenames of inner files, within each ZIP."
    )
    parser.add_argument(
        "--nprocesses", type=int, default=multiprocessing.cpu_count(),
        help="Specify the number of processes to run in parallel."
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Verbose output"
    )
    args = parser.parse_args()
    main_only_quicksetup_rootlogger(
        level=logging.DEBUG if args.verbose else logging.INFO)
    if args.files_with_matches and args.files_without_match:
        raise ValueError("Can't specify both --files_with_matches (-l) and "
                         "--files_without_match (-L)!")
    if bool(args.filenames_from_stdin) == bool(args.filename):
        raise ValueError("Specify --filenames_from_stdin or filenames on the "
                         "command line, but not both")

    # Compile regular expression
    if args.grep_inner_file_name:
        final_pattern = args.pattern
    else:
        encoding = getdefaultencoding()
        final_pattern = args.pattern.encode(encoding)
    flags = re.IGNORECASE if args.ignore_case else 0
    log.debug("Using regular expression {!r} with flags {!r}",
              final_pattern, flags)
    regex = re.compile(final_pattern, flags)

    # Set up pool for parallel processing
    pool = multiprocessing.Pool(processes=args.nprocesses)

    # Iterate through files
    parse_kwargs = dict(
        regex=regex,
        invert_match=args.invert_match,
        files_with_matches=args.files_with_matches,
        files_without_match=args.files_without_match,
        grep_inner_file_name=args.grep_inner_file_name,
        show_inner_file=args.show_inner_file
    )
    if args.filenames_from_stdin:
        for line in stdin.readlines():
            zipfilename = line.strip()
            parallel_kwargs = {'zipfilename': zipfilename}
            parallel_kwargs.update(**parse_kwargs)
            pool.apply_async(parse_zip, [], parallel_kwargs)
    else:
        for zipfilename in gen_filenames(starting_filenames=args.filename,
                                         recursive=args.recursive):
            parallel_kwargs = {'zipfilename': zipfilename}
            parallel_kwargs.update(**parse_kwargs)
            pool.apply_async(parse_zip, [], parallel_kwargs)
    pool.close()
    pool.join()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# setup.py

"""
cardinal_pythonlib setup file

To use:

    python setup.py sdist

    twine upload dist/*

To install in development mode:

    pip install -e .

"""

from setuptools import setup, find_packages
from codecs import open
from os import path

from cardinal_pythonlib.version_string import VERSION_STRING

PACKAGE_NAME = "cardinal_pythonlib"
THIS_DIR = path.abspath(path.dirname(__file__))
README_FILE = path.join(THIS_DIR, 'README.rst')  # read


# =============================================================================
# Get the long description from the README file
# =============================================================================

with open(README_FILE, encoding='utf-8') as f:
    long_description = f.read()


# =============================================================================
# Specify requirements
# =============================================================================

REQUIREMENTS = [
    # - Include most things that are imported without "try / except
    #   ImportError" handling.
    # - Include as few version requirements as possible.
    # - Keep it to pure-Python packages (for e.g. Windows installation with no
    #   compiler).
    # - Keep it to SMALL packages.
    # - SEE ALSO external_dependencies.rst

    "alembic",
    "appdirs>=1.4.0",
    "beautifulsoup4",  # "import bs4" or "from bs4 import ..."
    "colorlog",
    "isodate>=0.5.4",
    "numpy>=1.20.0",  # 1.20.0 required for numpy.typing
    "openpyxl",
    "pandas",
    "pendulum>=2.1.1",
    "prettytable",
    "psutil",
    "pygments",
    "pyparsing",
    "PyPDF2",
    "python-dateutil",  # "import dateutil"
    "scipy",
    "semantic-version",
    "SQLAlchemy",
    "sqlparse",
]

NOTES_RE_OTHER_REQUIREMENTS = """

# -----------------------------------------------------------------------------
# Use the PyPi index:
# -----------------------------------------------------------------------------
--index-url https://pypi.python.org/simple/


# -----------------------------------------------------------------------------
# The following are NOT HANDLED GRACEFULLY; their absence will cause a runtime
# ImportError, but we don't make them requirements as they need a compiler to
# install (and one might want to use the rest of the library without them).
# -----------------------------------------------------------------------------
# - SEE ALSO external_dependencies.rst

# arrow
# bcrypt
# colander
# deform
# Django>=2.0.0
# dogpile.cache
# pyramid
# webob  # installed by pyramid


# -----------------------------------------------------------------------------
# The following are OPTIONAL; their absence will be handled gracefully, so
# they are not requirements, but we note them here:
# -----------------------------------------------------------------------------
# - SEE ALSO external_dependencies.rst

# mmh3
# pdfkit
# pdfminer
# pypiwin32
# pyth
# python-docx   # "import docx"
# weasyprint
# xhtml2pdf


# -----------------------------------------------------------------------------
# FOR LIBRARY DEVELOPMENT
# -----------------------------------------------------------------------------

# sphinx
# sphinx_rtd_theme
# twine

# -----------------------------------------------------------------------------
# NO LONGER REQUIRED (but worth commenting on for now)
# -----------------------------------------------------------------------------

# DATABASE DRIVERS:
# jaydebeapi  -- in deprecated rnc_db module only
# mysql-python  # "import MySQLdb"  -- in deprecated rnc_db module only
# mysqlclient  # "import MySQLdb"  -- in deprecated rnc_db module only
# pymysql  -- in deprecated rnc_db module only
# pyodbc  -- in deprecated rnc_db module only
# pypyodbc  -- in deprecated rnc_db module only

"""


# =============================================================================
# setup args
# =============================================================================

setup(
    name=PACKAGE_NAME,

    version=VERSION_STRING,

    description='Miscellaneous Python libraries',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/RudolfCardinal/pythonlib',

    # Author details
    author='Rudolf Cardinal',
    author_email='rudolf@pobox.com',

    # Choose your license
    license='Apache License 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        'Natural Language :: English',

        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',

        'Topic :: Software Development :: Libraries',
    ],

    keywords='cardinal',

    packages=find_packages(),  # finds all the .py files in subdirectories

    install_requires=REQUIREMENTS,

    entry_points={
        'console_scripts': [
            # Format is 'script=module:function".

            # openxml
            'cardinalpythonlib_find_bad_openxml=cardinal_pythonlib.openxml.find_bad_openxml:main',  # noqa
            'cardinalpythonlib_find_recovered_openxml=cardinal_pythonlib.openxml.find_recovered_openxml:main',  # noqa
            'cardinalpythonlib_grep_in_openxml=cardinal_pythonlib.openxml.grep_in_openxml:main',  # noqa
            'cardinalpythonlib_pause_process_by_disk_space=cardinal_pythonlib.openxml.pause_process_by_disk_space:main',  # noqa

            # tools
            'cardinalpythonlib_backup_mysql_database=cardinal_pythonlib.tools.backup_mysql_database:main',  # noqa
            'cardinalpythonlib_bulk_email=cardinal_pythonlib.bulk_email.main:main',  # noqa
            'cardinalpythonlib_convert_athena_ohdsi_codes=cardinal_pythonlib.tools.convert_athena_ohdsi_codes:main',  # noqa
            'cardinalpythonlib_convert_mdb_to_mysql=cardinal_pythonlib.tools.convert_mdb_to_mysql:main',  # noqa
            'cardinalpythonlib_estimate_mysql_memory_usage=cardinal_pythonlib.tools.estimate_mysql_memory_usage:main',  # noqa
            'cardinalpythonlib_list_all_file_extensions=cardinal_pythonlib.tools.list_all_file_extensions:main',  # noqa
            'cardinalpythonlib_merge_csv=cardinal_pythonlib.tools.merge_csv:main',  # noqa
            'cardinalpythonlib_pdf_to_booklet=cardinal_pythonlib.tools.pdf_to_booklet:main',  # noqa
            'cardinalpythonlib_remove_duplicate_files=cardinal_pythonlib.tools.remove_duplicate_files:main',  # noqa

            # other
            'cardinalpythonlib_chebi=cardinal_pythonlib.chebi:main',
            'cardinalpythonlib_email=cardinal_pythonlib.email.sendmail:main',
            'cardinalpythonlib_extract_text=cardinal_pythonlib.extract_text:main',  # noqa
        ],
    },
)



#!/usr/bin/env python
# cardinal_pythonlib/typing_helpers.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Methods and unusual types for type hints.**

"""

from abc import abstractmethod
import csv
from typing import (
    Any,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TYPE_CHECKING,
    TypeVar,
    Union,
)


# =============================================================================
# Constants
# =============================================================================

T = TypeVar('T')


# =============================================================================
# with_typehint
# =============================================================================

def with_typehint(baseclass: Type[T]) -> Type[T]:
    """
    Useful function to make mixins with type hints from a base class.

    From
    https://stackoverflow.com/questions/51930339/how-do-i-correctly-add-type-hints-to-mixin-classes.

    Specimen usage:

    .. code-block:: python

        class MyMixin1(with_typehint(SomeBaseClass))):
            # ...

    """  # noqa
    if TYPE_CHECKING:
        return baseclass
    return object


def with_typehints(*baseclasses: Type[T]) -> Type[T]:
    """
    Useful function to make mixins with type hints from multiple base classes.

    From
    https://stackoverflow.com/questions/51930339/how-do-i-correctly-add-type-hints-to-mixin-classes.

    Specimen usage:

    .. code-block:: python

        class MyMixin2(*with_typehints(SomeBaseClass, AnotherBaseClass))):
            # ...

    """  # noqa
    if TYPE_CHECKING:
        return baseclasses
    return object


# =============================================================================
# csv.writer
# =============================================================================

class CSVWriterType(object):
    """
    Type hint for the result of ``csv.writer()``

    See https://stackoverflow.com/questions/51264355/how-to-type-annotate-object-returned-by-csv-writer
    """  # noqa

    @abstractmethod
    def writerow(self, row: List[str]) -> None:
        pass

    @abstractmethod
    def writerows(self, rows: List[List[str]]) -> None:
        pass

    @property
    @abstractmethod
    def dialect(self) -> csv.Dialect:
        pass


# =============================================================================
# Pep249DatabaseConnectionType
# =============================================================================

class Pep249DatabaseConnectionType(object):
    """
    Type hint for a database connection compliant with PEP 249. See
    https://www.python.org/dev/peps/pep-0249/.

    Not supported:

    - https://www.python.org/dev/peps/pep-0249/#optional-error-handling-extensions
    - https://www.python.org/dev/peps/pep-0249/#optional-two-phase-commit-extensions
    """  # noqa

    @abstractmethod
    def close(self) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#connection-objects
        """
        pass

    @abstractmethod
    def commit(self) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#connection-objects
        """
        pass

    @abstractmethod
    def rollback(self) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#connection-objects
        """
        pass

    @abstractmethod
    def cursor(self) -> "Pep249DatabaseCursorType":
        """
        See https://www.python.org/dev/peps/pep-0249/#connection-objects
        """
        pass

    @property
    @abstractmethod
    def messages(self) -> List[Tuple[Type, Any]]:
        """
        See
        https://www.python.org/dev/peps/pep-0249/#optional-db-api-extensions
        """
        pass


# =============================================================================
# Pep249DatabaseConnectionType
# =============================================================================

_DATABASE_ROW_TYPE = Sequence[Any]


class Pep249DatabaseCursorType(object):
    """
    Type hint for a database cursor compliant with PEP 249. See
    https://www.python.org/dev/peps/pep-0249/#cursor-objects

    Example, as per https://docs.python.org/3.6/library/sqlite3.html:

    .. code-block:: python

        import sqlite3
        conn = sqlite3.connect(':memory:')
        c = conn.cursor()
        c.execute('''
            CREATE TABLE stocks
            (date text, trans text, symbol text, qty real, price real)
        ''')
        c.execute('''
            INSERT INTO stocks VALUES ('2006-01-05','BUY','RHAT',100,35.14)
        ''')
        conn.commit()

        c.execute("SELECT * FROM stocks")
        print(repr(c.description))

        help(c)

    See also:

    - https://www.psycopg.org/docs/cursor.html

    """

    @abstractmethod
    def __init__(self, *args, **kwargs) -> None:
        pass

    @abstractmethod
    def __iter__(self) -> Iterator[_DATABASE_ROW_TYPE]:
        """
        See
        https://www.python.org/dev/peps/pep-0249/#optional-db-api-extensions
        """
        pass

    @abstractmethod
    def __new__(cls, *args, **kwargs) -> "Pep249DatabaseCursorType":
        pass

    @abstractmethod
    def __next__(self) -> None:
        pass

    @property
    @abstractmethod
    def description(self) \
            -> Optional[Sequence[Sequence[Any]]]:
        """
        A sequence of column_description objects, where each column_description
        describes one result column and has the following items:

        - name: ``str``
        - type_code: ``Optional[Type]``? Not sure.
        - display_size: ``Optional[int]``
        - internal_size: ``Optional[int]``
        - precision: ``Optional[int]``
        - scale: ``Optional[int]``
        - null_ok: ``Optional[bool]``

        The attribute is ``None`` for operations that don't return rows, and
        for un-executed cursors.

        """
        pass

    @property
    @abstractmethod
    def rowcount(self) -> int:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def callproc(self, procname: str, *args, **kwargs) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def execute(self, operation: str, *args, **kwargs) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def executemany(self, operation: str, *args, **kwargs) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def fetchone(self) -> Optional[_DATABASE_ROW_TYPE]:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def fetchmany(self, size: int = None) -> Sequence[_DATABASE_ROW_TYPE]:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def fetchall(self) -> Sequence[_DATABASE_ROW_TYPE]:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def nextset(self) -> Optional[bool]:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @property
    @abstractmethod
    def arraysize(self) -> int:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        # read/write attribute; see below
        pass

    # noinspection PyUnresolvedReferences
    @arraysize.setter
    @abstractmethod
    def arraysize(self, val: int) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        # https://stackoverflow.com/questions/35344209/python-abstract-property-setter-with-concrete-getter
        pass

    @abstractmethod
    def setinputsizes(self, sizes: Sequence[Union[Type, int]]) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @abstractmethod
    def setoutputsize(self, size: int, column: Optional[int]) -> None:
        """
        See https://www.python.org/dev/peps/pep-0249/#cursor-objects
        """
        pass

    @property
    @abstractmethod
    def connection(self) -> Pep249DatabaseConnectionType:
        """
        See
        https://www.python.org/dev/peps/pep-0249/#optional-db-api-extensions
        """
        pass

    @property
    @abstractmethod
    def lastrowid(self) -> Optional[int]:
        """
        See
        https://www.python.org/dev/peps/pep-0249/#optional-db-api-extensions
        """
        pass

    @property
    @abstractmethod
    def rownumber(self) -> Optional[int]:
        """
        See
        https://www.python.org/dev/peps/pep-0249/#optional-db-api-extensions
        """
        pass

    @abstractmethod
    def scroll(self, value: int, mode: str = 'relative') -> None:
        """
        See
        https://www.python.org/dev/peps/pep-0249/#optional-db-api-extensions
        """
        pass

    @property
    @abstractmethod
    def messages(self) -> List[Tuple[Type, Any]]:
        """
        See
        https://www.python.org/dev/peps/pep-0249/#optional-db-api-extensions
        """
        pass

    @abstractmethod
    def next(self) -> _DATABASE_ROW_TYPE:
        """
        See
        https://www.python.org/dev/peps/pep-0249/#optional-db-api-extensions
        """
        pass



#!/usr/bin/env python
# cardinal_pythonlib/dbfunc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to operate with the raw Python database API.**

See https://www.python.org/dev/peps/pep-0249/.

"""

from collections import OrderedDict
from typing import Any, Dict, Generator, List, Optional

from cardinal_pythonlib.typing_helpers import Pep249DatabaseCursorType as Cursor  # noqa


def get_fieldnames_from_cursor(cursor: Cursor) -> List[str]:
    """
    Get a list of fieldnames from an executed cursor.
    """
    return [i[0] for i in cursor.description]


def genrows(cursor: Cursor, arraysize: int = 1000) \
        -> Generator[List[Any], None, None]:
    """
    Generate all rows from a cursor.

    Args:
        cursor: the cursor
        arraysize: split fetches into chunks of this many records

    Yields:
        each row
    """
    # http://code.activestate.com/recipes/137270-use-generators-for-fetching-large-db-record-sets/  # noqa
    while True:
        results = cursor.fetchmany(arraysize)
        if not results:
            break
        for result in results:
            yield result


def genfirstvalues(cursor: Cursor, arraysize: int = 1000) \
        -> Generator[Any, None, None]:
    """
    Generate the first value in each row.

    Args:
        cursor: the cursor
        arraysize: split fetches into chunks of this many records

    Yields:
        the first value of each row
    """
    return (row[0] for row in genrows(cursor, arraysize))


def fetchallfirstvalues(cursor: Cursor) -> List[Any]:
    """
    Return a list of the first value in each row.
    """
    return [row[0] for row in cursor.fetchall()]


def gendicts(cursor: Cursor, arraysize: int = 1000) \
        -> Generator[Dict[str, Any], None, None]:
    """
    Generate all rows from a cursor as :class:`OrderedDict` objects.

    Args:
        cursor: the cursor
        arraysize: split fetches into chunks of this many records

    Yields:
        each row, as an :class:`OrderedDict` whose key are column names
        and whose values are the row values
    """
    columns = get_fieldnames_from_cursor(cursor)
    return (
        OrderedDict(zip(columns, row))
        for row in genrows(cursor, arraysize)
    )


def dictfetchall(cursor: Cursor) -> List[Dict[str, Any]]:
    """
    Return all rows from a cursor as a list of :class:`OrderedDict` objects.

    Args:
        cursor: the cursor

    Returns:
        a list (one item per row) of :class:`OrderedDict` objects whose key are
        column names and whose values are the row values
    """
    columns = get_fieldnames_from_cursor(cursor)
    return [
        OrderedDict(zip(columns, row))
        for row in cursor.fetchall()
    ]


def dictfetchone(cursor: Cursor) -> Optional[Dict[str, Any]]:
    """
    Return the next row from a cursor as an :class:`OrderedDict`, or ``None``.
    """
    columns = get_fieldnames_from_cursor(cursor)
    row = cursor.fetchone()
    if not row:
        return None
    return OrderedDict(zip(columns, row))



#!/usr/bin/env python
# cardinal_pythonlib/register_db_with_odbc.py

# noinspection HttpUrlsUsage
"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Creates/registers an Access database via ODBC.**

See

- https://support.microsoft.com/kb/126606/EN-US

  .. code-block:: none

    CREATE_DBV2=<path name> <sort order>
        (to create version 2 Jet engine mdb file, Access 2, 16bit)
    CREATE_DBV3=<path name> <sort order>
        (to create version 3 Jet engine mdb file, Access 95, Access 97)
    CREATE_DBV4=<path name> <sort order>
        (to create version 4 Jet engine mdb file, Access 2000)

- https://code.activestate.com/recipes/414879-create-an-odbc-data-source/

- view-source:https://www.experts-exchange.com/Programming/Languages/Pascal/Delphi/Q_22020226.html

  - this instead suggests CREATE_DBV3 for Access 95, CREATE_DBV4 for Access 97,
    and CREATE_DB for Access 2000, but that's probably wrong

- http://vieka.com/esqldoc/esqlref/htm/odbcsqlconfigdatasource.htm

- https://code.google.com/p/opendbviewer/source/browse/trunk/src/dbconnector/win32adodb.py?spec=svn45&r=45

- https://msdn.microsoft.com/en-us/library/aa140021(v=office.10).aspx

  - ... indicates that programmatic creation of queries/views (via ADO)
    leads to "invisible" queries

- https://msaccessmemento.hubpages.com/hub/Stored_Procedure_in_MS_Access

  - ... additionally, stored procedures have no user interface

"""  # noqa

from __future__ import division, print_function, absolute_import
import ctypes
import os.path
import platform
import sys
from typing import List

ODBC_ADD_DSN = 1         # Add data source
ODBC_CONFIG_DSN = 2      # Configure (edit) data source
ODBC_REMOVE_DSN = 3      # Remove data source
ODBC_ADD_SYS_DSN = 4     # add a system DSN
ODBC_CONFIG_SYS_DSN = 5  # Configure a system DSN
ODBC_REMOVE_SYS_DSN = 6  # remove a system DSN
access_driver = "Microsoft Access Driver (*.mdb)"
nul = chr(0)


def create_sys_dsn(driver: str, **kw) -> bool:
    """
    (Windows only.)
    Create a system ODBC data source name (DSN).

    Args:
        driver: ODBC driver name
        kw: Driver attributes

    Returns:
        bool: was the DSN created?
    """
    attributes = []  # type: List[str]
    for attr in kw.keys():
        attributes.append("%s=%s" % (attr, kw[attr]))
    return bool(
        ctypes.windll.ODBCCP32.SQLConfigDataSource(0, ODBC_ADD_SYS_DSN,
                                                   driver,
                                                   nul.join(attributes))
    )


def create_user_dsn(driver: str, **kw) -> bool:
    """
    (Windows only.)
    Create a user ODBC data source name (DSN).

    Args:
        driver: ODBC driver name
        kw: Driver attributes

    Returns:
        bool: was the DSN created?
    """
    attributes = []  # type: List[str]
    for attr in kw.keys():
        attributes.append("%s=%s" % (attr, kw[attr]))
    return bool(
        ctypes.windll.ODBCCP32.SQLConfigDataSource(0, ODBC_ADD_DSN, driver,
                                                   nul.join(attributes))
    )


def register_access_db(fullfilename: str, dsn: str, description: str) -> bool:
    """
    (Windows only.)
    Registers a Microsoft Access database with ODBC.

    Args:
        fullfilename: filename of the existing database
        dsn: ODBC data source name to create
        description: description of the database

    Returns:
        bool: was the DSN created?

    """
    directory = os.path.dirname(fullfilename)
    return create_sys_dsn(
        access_driver,
        SERVER="",
        DESCRIPTION=description,
        DSN=dsn,
        DBQ=fullfilename,
        DefaultDir=directory
    )


def create_and_register_access97_db(filename: str,
                                    dsn: str,
                                    description: str) -> bool:
    """
    (Windows only.)
    Creates a Microsoft Access 97 database and registers it with ODBC.

    Args:
        filename: filename of the database to create
        dsn: ODBC data source name to create
        description: description of the database

    Returns:
        bool: was the DSN created?
    """
    fullfilename = os.path.abspath(filename)
    create_string = fullfilename + " General"
    # ... filename, space, sort order ("General" for English)
    return (create_user_dsn(access_driver, CREATE_DB3=create_string) and
            register_access_db(filename, dsn, description))


def create_and_register_access2000_db(filename: str,
                                      dsn: str,
                                      description: str) -> bool:
    """
    (Windows only.)
    Creates a Microsoft Access 2000 database and registers it with ODBC.

    Args:
        filename: filename of the database to create
        dsn: ODBC data source name to create
        description: description of the database

    Returns:
        bool: was the DSN created?
    """
    fullfilename = os.path.abspath(filename)
    create_string = fullfilename + " General"
    # ... filename, space, sort order ("General" for English)
    return (create_user_dsn(access_driver, CREATE_DB4=create_string) and
            register_access_db(filename, dsn, description))


def create_and_register_access_db(filename: str,
                                  dsn: str,
                                  description: str) -> bool:
    """
    (Windows only.)
    Creates a Microsoft Access database and registers it with ODBC.

    Args:
        filename: filename of the database to create
        dsn: ODBC data source name to create
        description: description of the database

    Returns:
        bool: was the DSN created?
    """
    fullfilename = os.path.abspath(filename)
    create_string = fullfilename + " General"
    # ... filename, space, sort order ("General" for English)
    return (create_user_dsn(access_driver, CREATE_DB=create_string) and
            register_access_db(filename, dsn, description))
    # likely defaults to Access 2000


if __name__ == "__main__":
    if platform.system() != "Windows":
        print("Only Windows supported.")
        sys.exit()
    if create_and_register_access_db("testaccessdb.mdb",
                                     "Test_Access_DB",
                                     "My test Access DB DSN"):
        print("DSN created")
    else:
        print("DSN not created")



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/core_query.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Query helper functions using the SQLAlchemy Core.**

"""

from typing import Any, List, Optional, Sequence, Tuple, Union

from sqlalchemy.engine.base import Connection, Engine
from sqlalchemy.engine.result import ResultProxy
from sqlalchemy.orm.exc import MultipleResultsFound
from sqlalchemy.orm.session import Session
from sqlalchemy.sql.expression import (
    column, exists, func, literal, select, table,
)
from sqlalchemy.sql.schema import Table
from sqlalchemy.sql.selectable import Select

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.sqlalchemy.dialect import SqlaDialectName

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Get query result with fieldnames
# =============================================================================

def get_rows_fieldnames_from_raw_sql(
        session: Union[Session, Engine, Connection],
        sql: str) -> Tuple[Sequence[Sequence[Any]], Sequence[str]]:
    """
    Returns results and column names from a query.

    Args:
        session: SQLAlchemy :class:`Session`, :class:`Engine`, or
            :class:`Connection` object
        sql: raw SQL to execure

    Returns:
        ``(rows, fieldnames)`` where ``rows`` is the usual set of results and
        ``fieldnames`` are the name of the result columns/fields.

    """
    result = session.execute(sql)  # type: ResultProxy
    fieldnames = result.keys()
    rows = result.fetchall()
    return rows, fieldnames


# =============================================================================
# SELECT COUNT(*) (SQLAlchemy Core)
# =============================================================================
# https://stackoverflow.com/questions/12941416

def count_star(session: Union[Session, Engine, Connection],
               tablename: str,
               *criteria: Any) -> int:
    """
    Returns the result of ``COUNT(*)`` from the specified table (with
    additional ``WHERE`` criteria if desired).

    Args:
        session: SQLAlchemy :class:`Session`, :class:`Engine`, or
            :class:`Connection` object
        tablename: name of the table
        criteria: optional SQLAlchemy "where" criteria

    Returns:
        a scalar
    """
    # works if you pass a connection or a session or an engine; all have
    # the execute() method
    query = select([func.count()]).select_from(table(tablename))
    for criterion in criteria:
        query = query.where(criterion)
    return session.execute(query).scalar()


# =============================================================================
# SELECT COUNT(*), MAX(field) (SQLAlchemy Core)
# =============================================================================

def count_star_and_max(session: Union[Session, Engine, Connection],
                       tablename: str,
                       maxfield: str,
                       *criteria: Any) -> Tuple[int, Optional[int]]:
    """

    Args:
        session: SQLAlchemy :class:`Session`, :class:`Engine`, or
            :class:`Connection` object
        tablename: name of the table
        maxfield: name of column (field) to take the ``MAX()`` of
        criteria: optional SQLAlchemy "where" criteria

    Returns:
        a tuple: ``(count, maximum)``

    """
    query = select([
        func.count(),
        func.max(column(maxfield))
    ]).select_from(table(tablename))
    for criterion in criteria:
        query = query.where(criterion)
    result = session.execute(query)
    return result.fetchone()  # count, maximum


# =============================================================================
# SELECT EXISTS (SQLAlchemy Core)
# =============================================================================
# https://stackoverflow.com/questions/15381604
# http://docs.sqlalchemy.org/en/latest/orm/query.html

def exists_in_table(session: Session, table_: Table, *criteria: Any) -> bool:
    """
    Implements an efficient way of detecting if a record or records exist;
    should be faster than ``COUNT(*)`` in some circumstances.

    Args:
        session: SQLAlchemy :class:`Session`, :class:`Engine`, or
            :class:`Connection` object
        table_: SQLAlchemy :class:`Table` object
        criteria: optional SQLAlchemy "where" criteria

    Returns:
        a boolean

    Prototypical use:

    .. code-block:: python

        return exists_in_table(session,
                               table,
                               column(fieldname1) == value2,
                               column(fieldname2) == value2)
    """
    exists_clause = exists().select_from(table_)
    # ... EXISTS (SELECT * FROM tablename)
    for criterion in criteria:
        exists_clause = exists_clause.where(criterion)
    # ... EXISTS (SELECT * FROM tablename WHERE ...)

    if session.get_bind().dialect.name == SqlaDialectName.MSSQL:
        query = select([literal(True)]).where(exists_clause)
        # ... SELECT 1 WHERE EXISTS (SELECT * FROM tablename WHERE ...)
    else:
        query = select([exists_clause])
        # ... SELECT EXISTS (SELECT * FROM tablename WHERE ...)

    result = session.execute(query).scalar()
    return bool(result)


def exists_plain(session: Session, tablename: str, *criteria: Any) -> bool:
    """
    Implements an efficient way of detecting if a record or records exist;
    should be faster than COUNT(*) in some circumstances.

    Args:
        session: SQLAlchemy :class:`Session`, :class:`Engine`, or
            :class:`Connection` object
        tablename: name of the table
        criteria: optional SQLAlchemy "where" criteria

    Returns:
        a boolean

    Prototypical use:

    .. code-block:: python

        return exists_plain(config.destdb.session,
                            dest_table_name,
                            column(fieldname1) == value2,
                            column(fieldname2) == value2)
    """
    return exists_in_table(session, table(tablename), *criteria)


# =============================================================================
# Get all first values
# =============================================================================

def fetch_all_first_values(session: Session,
                           select_statement: Select) -> List[Any]:
    # noinspection HttpUrlsUsage
    """
    Returns a list of the first values in each row returned by a ``SELECT``
    query.

    A Core version of this sort of thing:
    http://xion.io/post/code/sqlalchemy-query-values.html

    Args:
        session: SQLAlchemy :class:`Session` object
        select_statement: SQLAlchemy :class:`Select` object

    Returns:
        a list of the first value of each result row

    """
    rows = session.execute(select_statement)  # type: ResultProxy
    try:
        return [row[0] for row in rows]
    except ValueError as e:
        raise MultipleResultsFound(str(e))



#!/usr/bin/env python
# cardinal_pythonlib/exceptions.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions for exception handling.**

"""

import sys
import traceback
from typing import Dict, NoReturn

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Exception handling
# =============================================================================

def add_info_to_exception(err: Exception, info: Dict) -> None:
    """
    Adds an information dictionary to an exception.

    See
    https://stackoverflow.com/questions/9157210/how-do-i-raise-the-same-exception-with-a-custom-message-in-python

    Args:
        err: the exception to be modified
        info: the information to add
    """  # noqa
    if not err.args:
        err.args = ('', )
    err.args += (info, )


def recover_info_from_exception(err: Exception) -> Dict:
    """
    Retrives the information added to an exception by
    :func:`add_info_to_exception`.
    """
    if len(err.args) < 1:
        return {}
    info = err.args[-1]
    if not isinstance(info, dict):
        return {}
    return info


def die(exc: Exception = None, exit_code: int = 1) -> NoReturn:
    """
    It is not clear that Python guarantees to exit with a non-zero exit code
    (errorlevel in DOS/Windows) upon an unhandled exception. So this function
    produces the usual stack trace then dies with the specified exit code.

    See
    https://stackoverflow.com/questions/9555133/e-printstacktrace-equivalent-in-python.

    Test code:

    .. code-block:: python

        import logging
        import sys
        import traceback
        logging.basicConfig(level=logging.DEBUG)
        log = logging.getLogger()

        def fail():
            try:
                x = 1/0
            except Exception as exc:
                die(exc)

    Then call

    .. code-block:: python

        fail()

    ... which should exit Python; then from Linux (for example):

    .. code-block:: bash

        echo $?  # show exit code

    """  # noqa
    if exc:
        lines = traceback.format_exception(
            None,  # etype: ignored
            exc,
            exc.__traceback__)  # https://www.python.org/dev/peps/pep-3134/
        msg = "".join(lines)
        # Method 1:
        # print("".join(lines), file=sys.stderr, flush=True)
        # Method 2:
        log.critical("{}", msg)
    log.critical("Exiting with exit code {}", exit_code)
    sys.exit(exit_code)



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/sqlfunc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to operate on SQL clauses for SQLAlchemy Core.**

"""

from typing import NoReturn, TYPE_CHECKING

from sqlalchemy.ext.compiler import compiles
# noinspection PyProtectedMember
from sqlalchemy.sql.expression import FunctionElement
from sqlalchemy.sql.sqltypes import Numeric

from cardinal_pythonlib.sqlalchemy.dialect import SqlaDialectName
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import ClauseElement, ClauseList
    from sqlalchemy.sql.compiler import SQLCompiler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Support functions
# =============================================================================

def fail_unknown_dialect(compiler: "SQLCompiler", task: str) -> NoReturn:
    """
    Raise :exc:`NotImplementedError` in relation to a dialect for which a
    function hasn't been implemented (with a helpful error message).
    """
    raise NotImplementedError(
        f"Don't know how to {task} on dialect {compiler.dialect!r}. "
        f"(Check also: if you printed the SQL before it was bound to an "
        f"engine, you will be trying to use a dialect like StrSQLCompiler, "
        f"which could be a reason for failure.)"
    )


def fetch_processed_single_clause(element: "ClauseElement",
                                  compiler: "SQLCompiler") -> str:
    """
    Takes a clause element that must have a single clause, and converts it
    to raw SQL text.
    """
    if len(element.clauses) != 1:
        raise TypeError(f"Only one argument supported; "
                        f"{len(element.clauses)} were passed")
    clauselist = element.clauses  # type: ClauseList
    first = clauselist.get_children()[0]
    return compiler.process(first)


# =============================================================================
# Extract year from a DATE/DATETIME etc.
# =============================================================================

# noinspection PyPep8Naming
class extract_year(FunctionElement):
    """
    Implements an SQLAlchemy :func:`extract_year` function, to extract the
    year from a date/datetime column.

    ``YEAR``, or ``func.year()``, is specific to some DBs, e.g. MySQL.
    So is ``EXTRACT``, or ``func.extract()``;
    https://modern-sql.com/feature/extract.

    This function therefore implements an ``extract_year`` function across
    multiple databases.

    Use this as:

    .. code-block:: python

      from cardinal_pythonlib.sqlalchemy.sqlfunc import extract_year

    ... then use :func:`extract_year` in an SQLAlchemy ``SELECT`` expression.

    Here's an example from CamCOPS:

    .. code-block:: python

        select_fields = [
            literal(cls.__tablename__).label("task"),
            extract_year(cls._when_added_batch_utc).label("year"),
            extract_month(cls._when_added_batch_utc).label("month"),
            func.count().label("num_tasks_added"),
        ]

    """
    type = Numeric()
    name = 'extract_year'


# noinspection PyUnusedLocal
@compiles(extract_year)
def extract_year_default(element: "ClauseElement",
                         compiler: "SQLCompiler", **kw) -> NoReturn:
    """
    Default implementation of :func:, which raises :exc:`NotImplementedError`.
    """
    fail_unknown_dialect(compiler, "extract year from date")


# noinspection PyUnusedLocal
@compiles(extract_year, SqlaDialectName.SQLSERVER)
@compiles(extract_year, SqlaDialectName.MYSQL)
def extract_year_year(element: "ClauseElement",
                      compiler: "SQLCompiler", **kw) -> str:
    # https://dev.mysql.com/doc/refman/5.5/en/date-and-time-functions.html#function_year  # noqa
    # https://docs.microsoft.com/en-us/sql/t-sql/functions/year-transact-sql
    clause = fetch_processed_single_clause(element, compiler)
    return f"YEAR({clause})"


# noinspection PyUnusedLocal
@compiles(extract_year, SqlaDialectName.ORACLE)
@compiles(extract_year, SqlaDialectName.POSTGRES)
def extract_year_extract(element: "ClauseElement",
                         compiler: "SQLCompiler", **kw) -> str:
    # https://docs.oracle.com/cd/B19306_01/server.102/b14200/functions050.htm
    clause = fetch_processed_single_clause(element, compiler)
    return f"EXTRACT(YEAR FROM {clause})"


# noinspection PyUnusedLocal
@compiles(extract_year, SqlaDialectName.SQLITE)
def extract_year_strftime(element: "ClauseElement",
                          compiler: "SQLCompiler", **kw) -> str:
    # https://sqlite.org/lang_datefunc.html
    clause = fetch_processed_single_clause(element, compiler)
    return f"STRFTIME('%Y', {clause})"


# =============================================================================
# Extract month from a DATE/DATETIME etc.
# =============================================================================

# noinspection PyPep8Naming
class extract_month(FunctionElement):
    """
    Implements an SQLAlchemy :func:`extract_month` function. See
    :func:`extract_year`.
    """
    type = Numeric()
    name = 'extract_month'


# noinspection PyUnusedLocal
@compiles(extract_month)
def extract_month_default(element: "ClauseElement",
                          compiler: "SQLCompiler", **kw) -> NoReturn:
    fail_unknown_dialect(compiler, "extract month from date")


# noinspection PyUnusedLocal
@compiles(extract_month, SqlaDialectName.SQLSERVER)
@compiles(extract_month, SqlaDialectName.MYSQL)
def extract_month_month(element: "ClauseElement",
                        compiler: "SQLCompiler", **kw) -> str:
    clause = fetch_processed_single_clause(element, compiler)
    return f"MONTH({clause})"


# noinspection PyUnusedLocal
@compiles(extract_month, SqlaDialectName.ORACLE)
@compiles(extract_year, SqlaDialectName.POSTGRES)
def extract_month_extract(element: "ClauseElement",
                          compiler: "SQLCompiler", **kw) -> str:
    clause = fetch_processed_single_clause(element, compiler)
    return f"EXTRACT(MONTH FROM {clause})"


# noinspection PyUnusedLocal
@compiles(extract_month, SqlaDialectName.SQLITE)
def extract_month_strftime(element: "ClauseElement",
                           compiler: "SQLCompiler", **kw) -> str:
    clause = fetch_processed_single_clause(element, compiler)
    return f"STRFTIME('%m', {clause})"


# =============================================================================
# Extract day (day of month, not day of week) from a DATE/DATETIME etc.
# =============================================================================

# noinspection PyPep8Naming
class extract_day_of_month(FunctionElement):
    """
    Implements an SQLAlchemy :func:`extract_day` function (to extract the day
    of the month from a date/datetime expression). See :func:`extract_year`.
    """
    type = Numeric()
    name = 'extract_day'


# noinspection PyUnusedLocal
@compiles(extract_day_of_month)
def extract_day_of_month_default(element: "ClauseElement",
                                 compiler: "SQLCompiler", **kw) -> NoReturn:
    fail_unknown_dialect(compiler, "extract day-of-month from date")


# noinspection PyUnusedLocal
@compiles(extract_day_of_month, SqlaDialectName.SQLSERVER)
@compiles(extract_day_of_month, SqlaDialectName.MYSQL)
def extract_day_of_month_day(element: "ClauseElement",
                             compiler: "SQLCompiler", **kw) -> str:
    clause = fetch_processed_single_clause(element, compiler)
    return f"DAY({clause})"


# noinspection PyUnusedLocal
@compiles(extract_day_of_month, SqlaDialectName.ORACLE)
@compiles(extract_year, SqlaDialectName.POSTGRES)
def extract_day_of_month_extract(element: "ClauseElement",
                                 compiler: "SQLCompiler", **kw) -> str:
    clause = fetch_processed_single_clause(element, compiler)
    return f"EXTRACT(DAY FROM {clause})"


# noinspection PyUnusedLocal
@compiles(extract_day_of_month, SqlaDialectName.SQLITE)
def extract_day_of_month_strftime(element: "ClauseElement",
                                  compiler: "SQLCompiler", **kw) -> str:
    clause = fetch_processed_single_clause(element, compiler)
    return f"STRFTIME('%d', {clause})"



#!/usr/bin/env python
# cardinal_pythonlib/rpm.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Randomized probability matching (RPM).**

As per:

- Scott SL. A modern Bayesian look at the multi-armed bandit.
  Applied Stochastic Models in Business and Industry 26 (2010): 639–58.
  https://doi.org/10.1002/asmb.874.

An R version is in ``rpm.R`` within https://github.com/rudolfcardinal/rlib.

"""

# =============================================================================
# Imports
# =============================================================================

import logging
import unittest

import numpy as np
import numpy.typing as npt
from scipy.integrate import quad
from scipy.stats import beta

log = logging.getLogger(__name__)


# =============================================================================
# RPM
# =============================================================================

def rpm_probabilities_successes_failures(
        n_successes: npt.ArrayLike,
        n_failures: npt.ArrayLike) -> np.ndarray:
    """
    Calculate the optimal choice probabilities.

    Note that Scott's original R version, compute.probopt(), on Figure 3 (p648)
    has arguments ``y`` (number of successes) and ``n`` (number of trials, NOT
    the number of failures).
    """
    k = len(n_successes)  # k is the number of actions
    assert len(n_failures) == k
    assert np.all(np.greater_equal(n_successes, 0))
    assert np.all(np.greater_equal(n_failures, 0))
    p_choice = np.zeros(shape=[k])  # answer to be populated

    n_successes_plus_one = np.array(n_successes) + 1
    n_failures_plus_one = np.array(n_failures) + 1
    for i in range(k):  # for each action:
        other_actions = list(range(0, i)) + list(range(i + 1, k))

        # Define function to integrate:
        def f(x: float) -> float:
            r = beta(n_successes_plus_one[i], n_failures_plus_one[i]).pdf(x)
            # ... beta density function, for the current action;
            # beta(a, b).pdf(x) is the probability density at x of the beta
            # distribution for a random variable with parameters a and b.
            # The R equivalent is dbeta(x, a, b).
            for j in other_actions:
                r *= beta(n_successes_plus_one[j],
                          n_failures_plus_one[j]).cdf(x)
                # ... for the other actions... beta(a, b).cdf(x) is the
                # cumulative distribution function of the beta distribution
                # with parameters a and b (the probability that a random
                # variable with parameters a,b is less than x). The R
                # equivalent is pbeta(x, a, b).
            return r

        # Integrate f from 0 to 1, e.g. via quad():
        # https://docs.scipy.org/doc/scipy/reference/integrate.html
        q = quad(f, 0, 1)[0]
        p_choice[i] = q

    return p_choice


def rpm_probabilities_successes_totals(
        n_successes: npt.ArrayLike,
        n_total: npt.ArrayLike) -> np.ndarray:
    """
    Randomized probability matching (RPM).

    Args:
        n_successes:
            Number of successes (per option).
        n_total:
            Total number of trials (per option).

    Returns:
        Optimal choice probabilities (per option) via RPM.
    """
    n_failures = np.array(n_total) - np.array(n_successes)
    return rpm_probabilities_successes_failures(n_successes, n_failures)


# =============================================================================
# Tests
# =============================================================================

class TestRpm(unittest.TestCase):
    def _assert_eq(self, name: str,
                   a1: npt.ArrayLike, a2: npt.ArrayLike,
                   abs_tolerance: float = 1e-8) -> None:
        self.assertTrue(
            np.allclose(a1, a2, atol=abs_tolerance),
            f"Error: {name}: {a1} != {a2}"
        )

    def test_rpm(self) -> None:
        p1 = rpm_probabilities_successes_failures(
            n_successes=[0, 0],
            n_failures=[0, 0]
        )
        self._assert_eq("p1", p1, [0.5, 0.5])

        p2 = rpm_probabilities_successes_totals(
            n_successes=[3, 7],
            n_total=[10, 10]
        )
        self._assert_eq("p2", p2, [0.04305447, 0.95694553])

        p3 = rpm_probabilities_successes_totals(
            n_successes=[1, 0],
            n_total=[1, 1]
        )
        self._assert_eq("p3", p3, [0.8333333, 0.1666667])

        self.assertRaises(
            AssertionError,
            rpm_probabilities_successes_totals,
            n_successes=[3, 7],
            n_total=[2, 10]
        )


# Test this file with:
# pytest rpm.py



#!/usr/bin/env python
# cardinal_pythonlib/debugging.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions for debugging.**

"""

import ctypes
import inspect
from inspect import FrameInfo
import logging
import pdb
import sys
import traceback
from types import FrameType
from typing import Any, Callable, List, Optional

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Debugging
# =============================================================================

def pdb_run(func: Callable, *args: Any, **kwargs: Any) -> Any:
    """
    Calls ``func(*args, **kwargs)``; if it raises an exception, break into
    the ``pdb`` debugger.
    """
    # noinspection PyBroadException
    try:
        return func(*args, **kwargs)
    except:  # nopep8
        type_, value, tb = sys.exc_info()
        traceback.print_exc()
        pdb.post_mortem(tb)


def cause_segfault() -> None:
    """
    This function will induce a segmentation fault and CRASH the application.
    Method as per https://docs.python.org/3/library/faulthandler.html
    """
    ctypes.string_at(0)  # will crash!


# =============================================================================
# Name of calling class/function, for status messages
# =============================================================================

def get_class_name_from_frame(fr: FrameType) -> Optional[str]:
    """
    A frame contains information about a specific call in the Python call
    stack; see https://docs.python.org/3/library/inspect.html.

    If the call was to a member function of a class, this function attempts
    to read the class's name. It returns ``None`` otherwise.
    """
    # https://stackoverflow.com/questions/2203424/python-how-to-retrieve-class-information-from-a-frame-object  # noqa
    args, _, _, value_dict = inspect.getargvalues(fr)
    # we check the first parameter for the frame function is named 'self'
    if len(args) and args[0] == 'self':
        # in that case, 'self' will be referenced in value_dict
        instance = value_dict.get('self', None)
        if instance:
            # return its class
            cls = getattr(instance, '__class__', None)
            if cls:
                return cls.__name__
            return None
    # return None otherwise
    return None


def get_caller_name(back: int = 0) -> str:
    """
    Return details about the CALLER OF THE CALLER (plus n calls further back)
    of this function.

    So, if your function calls :func:`get_caller_name`, it will return the
    name of the function that called your function! (Or ``back`` calls further
    back.)

    Example:

    .. code-block:: python

        from cardinal_pythonlib.debugging import get_caller_name

        def who_am_i():
            return get_caller_name()

        class MyClass(object):
            def classfunc(self):
                print("I am: " + who_am_i())
                print("I was called by: " + get_caller_name())
                print("That was called by: " + get_caller_name(back=1))

        def f2():
            x = MyClass()
            x.classfunc()

        def f1():
            f2()

        f1()

    will produce:

    .. code-block:: none

        I am: MyClass.classfunc
        I was called by: f2
        That was called by: f1

    """
    # https://stackoverflow.com/questions/5067604/determine-function-name-from-within-that-function-without-using-traceback  # noqa
    try:
        # noinspection PyProtectedMember
        frame = sys._getframe(back + 2)
    except ValueError:
        # Stack isn't deep enough.
        return '?'
    function_name = frame.f_code.co_name
    class_name = get_class_name_from_frame(frame)
    if class_name:
        return f"{class_name}.{function_name}"
    return function_name


# =============================================================================
# Who called us?
# =============================================================================

def get_caller_stack_info(start_back: int = 1) -> List[str]:
    r"""
    Retrieves a textual representation of the call stack.

    Args:
        start_back: number of calls back in the frame stack (starting
            from the frame stack as seen by :func:`get_caller_stack_info`)
            to begin with

    Returns:
        list of descriptions

    Example:

    .. code-block:: python

        from cardinal_pythonlib.debugging import get_caller_stack_info

        def who_am_i():
            return get_caller_name()

        class MyClass(object):
            def classfunc(self):
                print("Stack info:\n" + "\n".join(get_caller_stack_info()))

        def f2():
            x = MyClass()
            x.classfunc()

        def f1():
            f2()

        f1()

    if called from the Python prompt will produce:

    .. code-block:: none

        Stack info:
        <module>()
        ... defined at <stdin>:1
        ... line 1 calls next in stack; code is:

        f1()
        ... defined at <stdin>:1
        ... line 2 calls next in stack; code is:

        f2()
        ... defined at <stdin>:1
        ... line 3 calls next in stack; code is:

        classfunc(self=<__main__.MyClass object at 0x7f86a009c6d8>)
        ... defined at <stdin>:2
        ... line 3 calls next in stack; code is:

    and if called from a Python file will produce:

    .. code-block:: none

        Stack info:
        <module>()
        ... defined at /home/rudolf/tmp/stack.py:1
        ... line 17 calls next in stack; code is:
        f1()

        f1()
        ... defined at /home/rudolf/tmp/stack.py:14
        ... line 15 calls next in stack; code is:
            f2()

        f2()
        ... defined at /home/rudolf/tmp/stack.py:10
        ... line 12 calls next in stack; code is:
            x.classfunc()

        classfunc(self=<__main__.MyClass object at 0x7fd7a731f358>)
        ... defined at /home/rudolf/tmp/stack.py:7
        ... line 8 calls next in stack; code is:
                print("Stack info:\n" + "\n".join(get_caller_stack_info()))

    """
    # "0 back" is debug_callers, so "1 back" its caller
    # https://docs.python.org/3/library/inspect.html
    callers = []  # type: List[str]
    frameinfolist = inspect.stack()  # type: List[FrameInfo]  # noqa
    frameinfolist = frameinfolist[start_back:]
    for frameinfo in frameinfolist:
        frame = frameinfo.frame
        function_defined_at = "... defined at {filename}:{line}".format(
            filename=frame.f_code.co_filename,
            line=frame.f_code.co_firstlineno,
        )
        argvalues = inspect.getargvalues(frame)
        formatted_argvalues = inspect.formatargvalues(*argvalues)
        function_call = "{funcname}{argvals}".format(
            funcname=frame.f_code.co_name,
            argvals=formatted_argvalues,
        )
        code_context = frameinfo.code_context
        code = "".join(code_context) if code_context else ""
        onwards = "... line {line} calls next in stack; code is:\n{c}".format(
            line=frame.f_lineno,
            c=code,
        )
        description = "\n".join([function_call, function_defined_at, onwards])
        callers.append(description)
    return list(reversed(callers))


# =============================================================================
# Show the structure of an object in detail
# =============================================================================

def debug_object(obj, log_level: int = logging.DEBUG) -> None:
    """
    Sends details about a Python to the log, specifically its ``repr()``
    representation, and all of its attributes with their name, value, and type.

    Args:
        obj: object to debug
        log_level: log level to use; default is ``logging.DEBUG``
    """
    msgs = [f"For {obj!r}:"]
    for attrname in dir(obj):
        attribute = getattr(obj, attrname)
        msgs.append(
            f"- {attrname!r}: {attribute!r}, of type {type(attribute)!r}")
    log.log(log_level, "{}", "\n".join(msgs))



#!/usr/bin/env python
# cardinal_pythonlib/hash.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Hash functions**

In general, consider these hash functions:

- :func:`hash64`, using MurmurHash3 to provide a 64-bit integer: for fast
  INSECURE COMPARISON operations.
- an ``Hmac*`` class for SECURE cryptographic hashes.

Regarding None/NULL values (in CRATE):

- For difference detection, it may be helpful to be able to compare a standard
  hash, in which case ``somehash(None) == somehash("None") ==
  'abcdefsomething'``.

- It is vital not to hash NULL patient IDs, though: for example, two different
  patients without an NHS number must not be equated by comparison on a hash
  of the (NULL) NHS number.

- For anonymisation, this is handled in these functions:

  .. code-block:: none

    crate_anon/anonymise/anonymise.py / process_table()
    -> crate_anon/anonymise/configfiles.py / Config.encrypt_master_pid()
    -> crate_anon/anonymise/patient.py / Patient.get_rid
        ... via PatientInfo.rid
        ... to Config.encrypt_primary_pid()
"""

import hashlib
import hmac
import sys
from typing import Any, Callable, Tuple, Union

from sqlalchemy.sql.sqltypes import String, TypeEngine

try:
    # noinspection PyPackageRequirements
    import mmh3
except ImportError:
    mmh3 = None

# try:
#     import xxhash
#     pyhashxx = None
# except ImportError:
#     xxhash = None
#     import pyhashxx


# https://docs.python.org/3/library/platform.html#platform.architecture
IS_64_BIT = sys.maxsize > 2 ** 32
TIMING_HASH = "hash"


# =============================================================================
# Base classes
# =============================================================================

class GenericHasher(object):
    """
    Abstract base class for a hasher.
    """
    def hash(self, raw: Any) -> str:
        """
        Returns a hash of its input.
        """
        raise NotImplementedError()

    def output_length(self) -> int:
        """
        Returns the length of the hashes produced by this hasher.
        """
        return len(self.hash("dummytext"))

    def sqla_column_type(self) -> TypeEngine:
        """
        Returns a SQLAlchemy :class:`Column` type instance, specifically
        ``String(length=self.output_length())``.
        """
        return String(length=self.output_length())


# =============================================================================
# Simple salted hashers.
# =============================================================================

class GenericSaltedHasher(GenericHasher):
    """
    Generic representation of a simple salted hasher that stores a hash
    function and a salt.

    Note that these are vulnerable to attack: if an attacker knows a
    ``(message, digest)`` pair, it may be able to calculate another.
    See https://benlog.com/2008/06/19/dont-hash-secrets/ and
    https://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.134.8430

    **You should use HMAC instead if the thing you are hashing is secret.**
    """
    def __init__(self, hashfunc: Callable[[bytes], Any], salt: str) -> None:
        """
        Args:
            hashfunc: hash function to use
            salt: salt to use (following UTF-8 encoding)
        """
        self.hashfunc = hashfunc
        self.salt_bytes = salt.encode('utf-8')

    def hash(self, raw: Any) -> str:
        raw_bytes = str(raw).encode('utf-8')
        return self.hashfunc(self.salt_bytes + raw_bytes).hexdigest()


class MD5Hasher(GenericSaltedHasher):
    """
    Salted hasher based on MD5.

    MD5 is cryptographically FLAWED; avoid using it or this class.
    """
    def __init__(self, salt: str) -> None:
        super().__init__(hashlib.md5, salt)


class SHA256Hasher(GenericSaltedHasher):
    """
    Salted hasher based on SHA256.
    """
    def __init__(self, salt: str) -> None:
        super().__init__(hashlib.sha256, salt)


class SHA512Hasher(GenericSaltedHasher):
    """
    Salted hasher based on SHA512.
    """
    def __init__(self, salt: str) -> None:
        super().__init__(hashlib.sha512, salt)


# =============================================================================
# HMAC hashers. Better, if what you are hashing is secret.
# =============================================================================

class GenericHmacHasher(GenericHasher):
    """
    Generic representation of a hasher that hashes things via an HMAC
    (a hash-based message authentication code).
    See https://en.wikipedia.org/wiki/HMAC

    HMAC hashers are the thing to use if what you are hashing is secret.
    """
    def __init__(self, digestmod: Any, key: str) -> None:
        """
        Args:
            digestmod: see :func:`hmac.HMAC.__init__`
            key: cryptographic key to use
        """
        self.key_bytes = str(key).encode('utf-8')
        self.digestmod = digestmod

    def hash(self, raw: Any) -> str:
        """
        Returns the hex digest of a HMAC-encoded version of the input.
        """
        raw_bytes = str(raw).encode('utf-8')
        hmac_obj = hmac.new(key=self.key_bytes, msg=raw_bytes,
                            digestmod=self.digestmod)
        return hmac_obj.hexdigest()


class HmacMD5Hasher(GenericHmacHasher):
    """
    HMAC hasher based on MD5.
    (Even though MD5 is insecure, HMAC-MD5 is better. See Bellare M, Canetti R,
    Krawcyk H. Keying hash functions for message authentication. Lect. Notes
    Comput. Sci. Adv. Cryptol. - Crypto 96 Proc. 1996; 1109: 1–15.)
    """
    def __init__(self, key: str) -> None:
        super().__init__(hashlib.md5, key)


class HmacSHA256Hasher(GenericHmacHasher):
    """
    HMAC hasher based on SHA256.
    """
    def __init__(self, key: str) -> None:
        super().__init__(hashlib.sha256, key)


class HmacSHA512Hasher(GenericHmacHasher):
    """
    HMAC hasher based on SHA512.
    """
    def __init__(self, key: str) -> None:
        super().__init__(hashlib.sha512, key)


# =============================================================================
# Hash factory
# =============================================================================

class HashMethods(object):
    MD5 = "MD5"
    SHA256 = "SHA256"
    SHA512 = "SHA512"
    HMAC_MD5 = "HMAC_MD5"
    HMAC_SHA256 = "HMAC_SHA256"
    HMAC_SHA512 = "HMAC_SHA512"


def make_hasher(hash_method: str, key: str) -> GenericHasher:
    hash_method = hash_method.upper()
    if hash_method in (HashMethods.MD5, HashMethods.SHA256, HashMethods.SHA512):  # noqa
        raise ValueError(
            f"Non-HMAC hashers are deprecated for security reasons. You are "
            f"trying to use: {hash_method}")
    if hash_method == HashMethods.HMAC_MD5:
        return HmacMD5Hasher(key)
    elif hash_method == HashMethods.HMAC_SHA256 or not hash_method:
        return HmacSHA256Hasher(key)
    elif hash_method == HashMethods.HMAC_SHA512:
        return HmacSHA512Hasher(key)
    else:
        raise ValueError(f"Unknown value for hash_method: {hash_method}")


def get_longest_supported_hasher_output_length() -> int:
    dummyhash = make_hasher(HashMethods.HMAC_SHA512, "dummysalt")
    return dummyhash.output_length()


# =============================================================================
# Testing functions/notes relating to hashing
# =============================================================================

_ = """
import hashlib
from six.moves import range

class MD5Hasher(object):
    def __init__(self, salt):
        self.salt = salt
    def hash(self, raw):
        raw = str(raw)
        return hashlib.md5(self.salt + raw).hexdigest()

MAX_PID_STR = "9" * 10  # e.g. NHS numbers are 10-digit
MAX_PID_NUM = int(MAX_PID_STR)
# sets are MUCH, MUCH faster than lists for "have-I-seen-it" tests
hasher = MD5Hasher("dummysalt")
used_hashes = set()
for i in range(MAX_PID_NUM):
    if i % 1000000 == 0:
        print("... " + str(i))
    x = hasher.hash(i)
    if x in used_hashes:
        raise Exception("Collision! i={}".format(i))
    used_hashes.add(x)

# This gets increasingly slow but is certainly fine up to
#     282,000,000
# and we want to test
#   9,999,999,999
# Anyway, other people have done the work:
#   https://crypto.stackexchange.com/questions/15873
# ... and the value is expected to be at least 2^64, whereas an NHS number
# is less than 2^34 -- from math.log(9999999, 2).

"""


# =============================================================================
# Support functions
# =============================================================================

def to_bytes(data: Any) -> bytearray:
    """
    Convert anything to a ``bytearray``.

    See

    - https://stackoverflow.com/questions/7585435/best-way-to-convert-string-to-bytes-in-python-3
    - https://stackoverflow.com/questions/10459067/how-to-convert-my-bytearrayb-x9e-x18k-x9a-to-something-like-this-x9e-x11
    """  # noqa
    if isinstance(data, int):
        return bytearray([data])
    return bytearray(data, encoding='latin-1')


def to_str(data: Any) -> str:
    """
    Convert anything to a ``str``.
    """
    return str(data)


def twos_comp_to_signed(val: int, n_bits: int) -> int:
    """
    Convert a "two's complement" representation (as an integer) to its signed
    version.

    Args:
        val: positive integer representing a number in two's complement format
        n_bits: number of bits (which must reflect a whole number of bytes)

    Returns:
        signed integer

    See https://stackoverflow.com/questions/1604464/twos-complement-in-python

    """
    assert n_bits % 8 == 0, "Must specify a whole number of bytes"
    n_bytes = n_bits // 8
    b = val.to_bytes(n_bytes, byteorder=sys.byteorder, signed=False)
    return int.from_bytes(b, byteorder=sys.byteorder, signed=True)


def signed_to_twos_comp(val: int, n_bits: int) -> int:
    """
    Convert a signed integer to its "two's complement" representation.

    Args:
        val: signed integer
        n_bits: number of bits (which must reflect a whole number of bytes)

    Returns:
        unsigned integer: two's complement version

    """
    assert n_bits % 8 == 0, "Must specify a whole number of bytes"
    n_bytes = n_bits // 8
    b = val.to_bytes(n_bytes, byteorder=sys.byteorder, signed=True)
    return int.from_bytes(b, byteorder=sys.byteorder, signed=False)


def bytes_to_long(bytesdata: bytes) -> int:
    """
    Converts an 8-byte sequence to a long integer.

    Args:
        bytesdata: 8 consecutive bytes, as a ``bytes`` object, in
            little-endian format (least significant byte [LSB] first)

    Returns:
        integer

    """
    assert len(bytesdata) == 8
    return sum((b << (k * 8) for k, b in enumerate(bytesdata)))


# =============================================================================
# Pure Python implementations of MurmurHash3
# =============================================================================

# -----------------------------------------------------------------------------
# SO ones
# -----------------------------------------------------------------------------

def murmur3_x86_32(data: Union[bytes, bytearray], seed: int = 0) -> int:
    """
    Pure 32-bit Python implementation of MurmurHash3; see
    https://stackoverflow.com/questions/13305290/is-there-a-pure-python-implementation-of-murmurhash.

    Args:
        data: data to hash
        seed: seed

    Returns:
        integer hash

    """  # noqa
    c1 = 0xcc9e2d51
    c2 = 0x1b873593

    length = len(data)
    h1 = seed
    rounded_end = (length & 0xfffffffc)  # round down to 4 byte block
    for i in range(0, rounded_end, 4):
        # little endian load order
        # RNC: removed ord() calls
        k1 = (data[i] & 0xff) | ((data[i + 1] & 0xff) << 8) | \
             ((data[i + 2] & 0xff) << 16) | (data[i + 3] << 24)
        k1 *= c1
        k1 = (k1 << 15) | ((k1 & 0xffffffff) >> 17)  # ROTL32(k1, 15)
        k1 *= c2

        h1 ^= k1
        h1 = (h1 << 13) | ((h1 & 0xffffffff) >> 19)  # ROTL32(h1, 13)
        h1 = h1 * 5 + 0xe6546b64

    # tail
    k1 = 0

    val = length & 0x03
    if val == 3:
        k1 = (data[rounded_end + 2] & 0xff) << 16
    # fallthrough
    if val in (2, 3):
        k1 |= (data[rounded_end + 1] & 0xff) << 8
    # fallthrough
    if val in (1, 2, 3):
        k1 |= data[rounded_end] & 0xff
        k1 *= c1
        k1 = (k1 << 15) | ((k1 & 0xffffffff) >> 17)  # ROTL32(k1, 15)
        k1 *= c2
        h1 ^= k1

    # finalization
    h1 ^= length

    # fmix(h1)
    h1 ^= ((h1 & 0xffffffff) >> 16)
    h1 *= 0x85ebca6b
    h1 ^= ((h1 & 0xffffffff) >> 13)
    h1 *= 0xc2b2ae35
    h1 ^= ((h1 & 0xffffffff) >> 16)

    return h1 & 0xffffffff


# noinspection PyPep8
def murmur3_64(data: Union[bytes, bytearray], seed: int = 19820125) -> int:
    """
    Pure 64-bit Python implementation of MurmurHash3; see
    https://stackoverflow.com/questions/13305290/is-there-a-pure-python-implementation-of-murmurhash
    (plus RNC bugfixes).

    Args:
        data: data to hash
        seed: seed

    Returns:
        integer hash
    """  # noqa
    m = 0xc6a4a7935bd1e995
    r = 47

    mask = 2 ** 64 - 1

    length = len(data)

    h = seed ^ ((m * length) & mask)

    offset = (length // 8) * 8
    # RNC: was /, but for Python 3 that gives float; brackets added for clarity
    for ll in range(0, offset, 8):
        k = bytes_to_long(data[ll:ll + 8])
        k = (k * m) & mask
        k ^= (k >> r) & mask
        k = (k * m) & mask
        h = (h ^ k)
        h = (h * m) & mask

    # Variable was named "l"; renamed to "l_" for PEP8
    l_ = length & 7

    if l_ >= 7:
        h = (h ^ (data[offset + 6] << 48))

    if l_ >= 6:
        h = (h ^ (data[offset + 5] << 40))

    if l_ >= 5:
        h = (h ^ (data[offset + 4] << 32))

    if l_ >= 4:
        h = (h ^ (data[offset + 3] << 24))

    if l_ >= 3:
        h = (h ^ (data[offset + 2] << 16))

    if l_ >= 2:
        h = (h ^ (data[offset + 1] << 8))

    if l_ >= 1:
        h = (h ^ data[offset])
        h = (h * m) & mask

    h ^= (h >> r) & mask
    h = (h * m) & mask
    h ^= (h >> r) & mask

    return h


# -----------------------------------------------------------------------------
# pymmh3 ones, renamed, with some bugfixes
# -----------------------------------------------------------------------------

def pymmh3_hash128_x64(key: Union[bytes, bytearray], seed: int) -> int:
    """
    Implements 128-bit murmur3 hash for x64, as per ``pymmh3``, with some
    bugfixes.

    Args:
        key: data to hash
        seed: seed

    Returns:
        integer hash
    """

    def fmix(k):
        k ^= k >> 33
        k = (k * 0xff51afd7ed558ccd) & 0xFFFFFFFFFFFFFFFF
        k ^= k >> 33
        k = (k * 0xc4ceb9fe1a85ec53) & 0xFFFFFFFFFFFFFFFF
        k ^= k >> 33
        return k

    length = len(key)
    nblocks = int(length / 16)

    h1 = seed
    h2 = seed

    c1 = 0x87c37b91114253d5
    c2 = 0x4cf5ad432745937f

    # body
    for block_start in range(0, nblocks * 8, 8):
        # ??? big endian?
        k1 = (
            key[2 * block_start + 7] << 56 |
            key[2 * block_start + 6] << 48 |
            key[2 * block_start + 5] << 40 |
            key[2 * block_start + 4] << 32 |
            key[2 * block_start + 3] << 24 |
            key[2 * block_start + 2] << 16 |
            key[2 * block_start + 1] << 8 |
            key[2 * block_start + 0]
        )

        k2 = (
            key[2 * block_start + 15] << 56 |
            key[2 * block_start + 14] << 48 |
            key[2 * block_start + 13] << 40 |
            key[2 * block_start + 12] << 32 |
            key[2 * block_start + 11] << 24 |
            key[2 * block_start + 10] << 16 |
            key[2 * block_start + 9] << 8 |
            key[2 * block_start + 8]
        )

        k1 = (c1 * k1) & 0xFFFFFFFFFFFFFFFF
        k1 = (k1 << 31 | k1 >> 33) & 0xFFFFFFFFFFFFFFFF  # inlined ROTL64
        k1 = (c2 * k1) & 0xFFFFFFFFFFFFFFFF
        h1 ^= k1

        h1 = (h1 << 27 | h1 >> 37) & 0xFFFFFFFFFFFFFFFF  # inlined ROTL64
        h1 = (h1 + h2) & 0xFFFFFFFFFFFFFFFF
        h1 = (h1 * 5 + 0x52dce729) & 0xFFFFFFFFFFFFFFFF

        k2 = (c2 * k2) & 0xFFFFFFFFFFFFFFFF
        k2 = (k2 << 33 | k2 >> 31) & 0xFFFFFFFFFFFFFFFF  # inlined ROTL64
        k2 = (c1 * k2) & 0xFFFFFFFFFFFFFFFF
        h2 ^= k2

        h2 = (h2 << 31 | h2 >> 33) & 0xFFFFFFFFFFFFFFFF  # inlined ROTL64
        h2 = (h1 + h2) & 0xFFFFFFFFFFFFFFFF
        h2 = (h2 * 5 + 0x38495ab5) & 0xFFFFFFFFFFFFFFFF

    # tail
    tail_index = nblocks * 16
    k1 = 0
    k2 = 0
    tail_size = length & 15

    if tail_size >= 15:
        k2 ^= key[tail_index + 14] << 48
    if tail_size >= 14:
        k2 ^= key[tail_index + 13] << 40
    if tail_size >= 13:
        k2 ^= key[tail_index + 12] << 32
    if tail_size >= 12:
        k2 ^= key[tail_index + 11] << 24
    if tail_size >= 11:
        k2 ^= key[tail_index + 10] << 16
    if tail_size >= 10:
        k2 ^= key[tail_index + 9] << 8
    if tail_size >= 9:
        k2 ^= key[tail_index + 8]

    if tail_size > 8:
        k2 = (k2 * c2) & 0xFFFFFFFFFFFFFFFF
        k2 = (k2 << 33 | k2 >> 31) & 0xFFFFFFFFFFFFFFFF  # inlined ROTL64
        k2 = (k2 * c1) & 0xFFFFFFFFFFFFFFFF
        h2 ^= k2

    if tail_size >= 8:
        k1 ^= key[tail_index + 7] << 56
    if tail_size >= 7:
        k1 ^= key[tail_index + 6] << 48
    if tail_size >= 6:
        k1 ^= key[tail_index + 5] << 40
    if tail_size >= 5:
        k1 ^= key[tail_index + 4] << 32
    if tail_size >= 4:
        k1 ^= key[tail_index + 3] << 24
    if tail_size >= 3:
        k1 ^= key[tail_index + 2] << 16
    if tail_size >= 2:
        k1 ^= key[tail_index + 1] << 8
    if tail_size >= 1:
        k1 ^= key[tail_index + 0]

    if tail_size > 0:
        k1 = (k1 * c1) & 0xFFFFFFFFFFFFFFFF
        k1 = (k1 << 31 | k1 >> 33) & 0xFFFFFFFFFFFFFFFF  # inlined ROTL64
        k1 = (k1 * c2) & 0xFFFFFFFFFFFFFFFF
        h1 ^= k1

    # finalization
    h1 ^= length
    h2 ^= length

    h1 = (h1 + h2) & 0xFFFFFFFFFFFFFFFF
    h2 = (h1 + h2) & 0xFFFFFFFFFFFFFFFF

    h1 = fmix(h1)
    h2 = fmix(h2)

    h1 = (h1 + h2) & 0xFFFFFFFFFFFFFFFF
    h2 = (h1 + h2) & 0xFFFFFFFFFFFFFFFF

    return h2 << 64 | h1


def pymmh3_hash128_x86(key: Union[bytes, bytearray], seed: int) -> int:
    """
    Implements 128-bit murmur3 hash for x86, as per ``pymmh3``, with some
    bugfixes.

    Args:
        key: data to hash
        seed: seed

    Returns:
        integer hash
    """

    def fmix(h):
        h ^= h >> 16
        h = (h * 0x85ebca6b) & 0xFFFFFFFF
        h ^= h >> 13
        h = (h * 0xc2b2ae35) & 0xFFFFFFFF
        h ^= h >> 16
        return h

    length = len(key)
    nblocks = int(length / 16)

    h1 = seed
    h2 = seed
    h3 = seed
    h4 = seed

    c1 = 0x239b961b
    c2 = 0xab0e9789
    c3 = 0x38b34ae5
    c4 = 0xa1e38b93

    # body
    for block_start in range(0, nblocks * 16, 16):
        k1 = (
            key[block_start + 3] << 24 |
            key[block_start + 2] << 16 |
            key[block_start + 1] << 8 |
            key[block_start + 0]
        )
        k2 = (
            key[block_start + 7] << 24 |
            key[block_start + 6] << 16 |
            key[block_start + 5] << 8 |
            key[block_start + 4]
        )
        k3 = (
            key[block_start + 11] << 24 |
            key[block_start + 10] << 16 |
            key[block_start + 9] << 8 |
            key[block_start + 8]
        )
        k4 = (
            key[block_start + 15] << 24 |
            key[block_start + 14] << 16 |
            key[block_start + 13] << 8 |
            key[block_start + 12]
        )

        k1 = (c1 * k1) & 0xFFFFFFFF
        k1 = (k1 << 15 | k1 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        k1 = (c2 * k1) & 0xFFFFFFFF
        h1 ^= k1

        h1 = (h1 << 19 | h1 >> 13) & 0xFFFFFFFF  # inlined ROTL32
        h1 = (h1 + h2) & 0xFFFFFFFF
        h1 = (h1 * 5 + 0x561ccd1b) & 0xFFFFFFFF

        k2 = (c2 * k2) & 0xFFFFFFFF
        k2 = (k2 << 16 | k2 >> 16) & 0xFFFFFFFF  # inlined ROTL32
        k2 = (c3 * k2) & 0xFFFFFFFF
        h2 ^= k2

        h2 = (h2 << 17 | h2 >> 15) & 0xFFFFFFFF  # inlined ROTL32
        h2 = (h2 + h3) & 0xFFFFFFFF
        h2 = (h2 * 5 + 0x0bcaa747) & 0xFFFFFFFF

        k3 = (c3 * k3) & 0xFFFFFFFF
        k3 = (k3 << 17 | k3 >> 15) & 0xFFFFFFFF  # inlined ROTL32
        k3 = (c4 * k3) & 0xFFFFFFFF
        h3 ^= k3

        h3 = (h3 << 15 | h3 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        h3 = (h3 + h4) & 0xFFFFFFFF
        h3 = (h3 * 5 + 0x96cd1c35) & 0xFFFFFFFF

        k4 = (c4 * k4) & 0xFFFFFFFF
        k4 = (k4 << 18 | k4 >> 14) & 0xFFFFFFFF  # inlined ROTL32
        k4 = (c1 * k4) & 0xFFFFFFFF
        h4 ^= k4

        h4 = (h4 << 13 | h4 >> 19) & 0xFFFFFFFF  # inlined ROTL32
        h4 = (h1 + h4) & 0xFFFFFFFF
        h4 = (h4 * 5 + 0x32ac3b17) & 0xFFFFFFFF

    # tail
    tail_index = nblocks * 16
    k1 = 0
    k2 = 0
    k3 = 0
    k4 = 0
    tail_size = length & 15

    if tail_size >= 15:
        k4 ^= key[tail_index + 14] << 16
    if tail_size >= 14:
        k4 ^= key[tail_index + 13] << 8
    if tail_size >= 13:
        k4 ^= key[tail_index + 12]

    if tail_size > 12:
        k4 = (k4 * c4) & 0xFFFFFFFF
        k4 = (k4 << 18 | k4 >> 14) & 0xFFFFFFFF  # inlined ROTL32
        k4 = (k4 * c1) & 0xFFFFFFFF
        h4 ^= k4

    if tail_size >= 12:
        k3 ^= key[tail_index + 11] << 24
    if tail_size >= 11:
        k3 ^= key[tail_index + 10] << 16
    if tail_size >= 10:
        k3 ^= key[tail_index + 9] << 8
    if tail_size >= 9:
        k3 ^= key[tail_index + 8]

    if tail_size > 8:
        k3 = (k3 * c3) & 0xFFFFFFFF
        k3 = (k3 << 17 | k3 >> 15) & 0xFFFFFFFF  # inlined ROTL32
        k3 = (k3 * c4) & 0xFFFFFFFF
        h3 ^= k3

    if tail_size >= 8:
        k2 ^= key[tail_index + 7] << 24
    if tail_size >= 7:
        k2 ^= key[tail_index + 6] << 16
    if tail_size >= 6:
        k2 ^= key[tail_index + 5] << 8
    if tail_size >= 5:
        k2 ^= key[tail_index + 4]

    if tail_size > 4:
        k2 = (k2 * c2) & 0xFFFFFFFF
        k2 = (k2 << 16 | k2 >> 16) & 0xFFFFFFFF  # inlined ROTL32
        k2 = (k2 * c3) & 0xFFFFFFFF
        h2 ^= k2

    if tail_size >= 4:
        k1 ^= key[tail_index + 3] << 24
    if tail_size >= 3:
        k1 ^= key[tail_index + 2] << 16
    if tail_size >= 2:
        k1 ^= key[tail_index + 1] << 8
    if tail_size >= 1:
        k1 ^= key[tail_index + 0]

    if tail_size > 0:
        k1 = (k1 * c1) & 0xFFFFFFFF
        k1 = (k1 << 15 | k1 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        k1 = (k1 * c2) & 0xFFFFFFFF
        h1 ^= k1

    # finalization
    h1 ^= length
    h2 ^= length
    h3 ^= length
    h4 ^= length

    h1 = (h1 + h2) & 0xFFFFFFFF
    h1 = (h1 + h3) & 0xFFFFFFFF
    h1 = (h1 + h4) & 0xFFFFFFFF
    h2 = (h1 + h2) & 0xFFFFFFFF
    h3 = (h1 + h3) & 0xFFFFFFFF
    h4 = (h1 + h4) & 0xFFFFFFFF

    h1 = fmix(h1)
    h2 = fmix(h2)
    h3 = fmix(h3)
    h4 = fmix(h4)

    h1 = (h1 + h2) & 0xFFFFFFFF
    h1 = (h1 + h3) & 0xFFFFFFFF
    h1 = (h1 + h4) & 0xFFFFFFFF
    h2 = (h1 + h2) & 0xFFFFFFFF
    h3 = (h1 + h3) & 0xFFFFFFFF
    h4 = (h1 + h4) & 0xFFFFFFFF

    return h4 << 96 | h3 << 64 | h2 << 32 | h1


def pymmh3_hash128(key: Union[bytes, bytearray],
                   seed: int = 0,
                   x64arch: bool = True) -> int:
    """
    Implements 128bit murmur3 hash, as per ``pymmh3``.

    Args:
        key: data to hash
        seed: seed
        x64arch: is a 64-bit architecture available?

    Returns:
        integer hash

    """
    if x64arch:
        return pymmh3_hash128_x64(key, seed)
    else:
        return pymmh3_hash128_x86(key, seed)


def pymmh3_hash64(key: Union[bytes, bytearray],
                  seed: int = 0,
                  x64arch: bool = True) -> Tuple[int, int]:
    """
    Implements 64bit murmur3 hash, as per ``pymmh3``. Returns a tuple.

    Args:
        key: data to hash
        seed: seed
        x64arch: is a 64-bit architecture available?

    Returns:
        tuple: tuple of integers, ``(signed_val1, signed_val2)``

    """

    hash_128 = pymmh3_hash128(key, seed, x64arch)

    unsigned_val1 = hash_128 & 0xFFFFFFFFFFFFFFFF  # low half
    if unsigned_val1 & 0x8000000000000000 == 0:
        signed_val1 = unsigned_val1
    else:
        signed_val1 = -((unsigned_val1 ^ 0xFFFFFFFFFFFFFFFF) + 1)

    unsigned_val2 = (hash_128 >> 64) & 0xFFFFFFFFFFFFFFFF  # high half
    if unsigned_val2 & 0x8000000000000000 == 0:
        signed_val2 = unsigned_val2
    else:
        signed_val2 = -((unsigned_val2 ^ 0xFFFFFFFFFFFFFFFF) + 1)

    return signed_val1, signed_val2


# =============================================================================
# Checks
# =============================================================================

def compare_python_to_reference_murmur3_32(data: Any, seed: int = 0) -> None:
    """
    Checks the pure Python implementation of 32-bit murmur3 against the
    ``mmh3`` C-based module.

    Args:
        data: data to hash
        seed: seed

    Raises:
        AssertionError: if the two calculations don't match

    """
    assert mmh3, "Need mmh3 module"
    c_data = to_str(data)
    # noinspection PyUnresolvedReferences
    c_signed = mmh3.hash(c_data, seed=seed)  # 32 bit
    py_data = to_bytes(c_data)
    py_unsigned = murmur3_x86_32(py_data, seed=seed)
    py_signed = twos_comp_to_signed(py_unsigned, n_bits=32)
    preamble = f"Hashing {data!r} with MurmurHash3/32-bit/seed={seed}"
    if c_signed == py_signed:
        print(preamble + f" -> {c_signed}: OK")
    else:
        raise AssertionError(
            preamble +
            f"; mmh3 says {c_data!r} -> {c_signed}, "
            f"Python version says {py_data!r} -> {py_unsigned} = {py_signed}")


def compare_python_to_reference_murmur3_64(data: Any, seed: int = 0) -> None:
    """
    Checks the pure Python implementation of 64-bit murmur3 against the
    ``mmh3`` C-based module.

    Args:
        data: data to hash
        seed: seed

    Raises:
        AssertionError: if the two calculations don't match

    """
    assert mmh3, "Need mmh3 module"
    c_data = to_str(data)
    # noinspection PyUnresolvedReferences
    c_signed_low, c_signed_high = mmh3.hash64(c_data, seed=seed,
                                              x64arch=IS_64_BIT)
    py_data = to_bytes(c_data)
    py_signed_low, py_signed_high = pymmh3_hash64(py_data, seed=seed)
    preamble = (
        f"Hashing {data!r} with MurmurHash3/64-bit values from 128-bit "
        f"hash/seed={seed}"
    )
    if c_signed_low == py_signed_low and c_signed_high == py_signed_high:
        print(preamble + f" -> (low={c_signed_low}, high={c_signed_high}): OK")
    else:
        raise AssertionError(
            preamble +
            f"; mmh3 says {c_data!r} -> "
            f"(low={c_signed_low}, high={c_signed_high}), "
            f"Python version says {py_data!r} -> "
            f"(low={py_signed_low}, high={py_signed_high})")


# =============================================================================
# Hashing in a NON-CRYPTOGRAPHIC, PREDICTABLE, and fast way
# =============================================================================

def hash32(data: Any, seed: int = 0) -> int:
    """
    Non-cryptographic, deterministic, fast hash.

    Args:
        data: data to hash
        seed: seed

    Returns:
        signed 32-bit integer
    """
    c_data = to_str(data)
    if mmh3:
        # noinspection PyUnresolvedReferences
        return mmh3.hash(c_data, seed=seed)
    py_data = to_bytes(c_data)
    py_unsigned = murmur3_x86_32(py_data, seed=seed)
    return twos_comp_to_signed(py_unsigned, n_bits=32)


def hash64(data: Any, seed: int = 0) -> int:
    """
    Non-cryptographic, deterministic, fast hash.

    Args:
        data: data to hash
        seed: seed

    Returns:
        signed 64-bit integer
    """
    # -------------------------------------------------------------------------
    # MurmurHash3
    # -------------------------------------------------------------------------
    c_data = to_str(data)
    if mmh3:
        # noinspection PyUnresolvedReferences
        c_signed_low, _ = mmh3.hash64(data, seed=seed, x64arch=IS_64_BIT)
        return c_signed_low
    py_data = to_bytes(c_data)
    py_signed_low, _ = pymmh3_hash64(py_data, seed=seed)
    return py_signed_low

    # -------------------------------------------------------------------------
    # xxHash
    # -------------------------------------------------------------------------
    # if xxhash:
    #     hasher = xxhash.xxh64(seed=0)
    #     hasher.update(data)
    #     return hasher.intdigest()
    # else:
    #     hasher = pyhashxx.Hashxx(seed=0)
    #     # then do some update, but it doesn't like plain strings...
    #     return hasher.digest()


# =============================================================================
# Testing
# =============================================================================

def main() -> None:
    """
    Command-line validation checks.
    """
    _ = """
    print(twos_comp_to_signed(0, n_bits=32))  # 0
    print(twos_comp_to_signed(2 ** 31 - 1, n_bits=32))  # 2147483647
    print(twos_comp_to_signed(2 ** 31, n_bits=32))  # -2147483648 == -(2 ** 31)
    print(twos_comp_to_signed(2 ** 32 - 1, n_bits=32))  # -1
    print(signed_to_twos_comp(-1, n_bits=32))  # 4294967295 = 2 ** 32 - 1
    print(signed_to_twos_comp(-(2 ** 31), n_bits=32))  # 2147483648 = 2 ** 31 - 1
    """  # noqa
    testdata = [
        "hello",
        1,
        ["bongos", "today"],
    ]
    for data in testdata:
        compare_python_to_reference_murmur3_32(data, seed=0)
        compare_python_to_reference_murmur3_64(data, seed=0)
    print("All OK")


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# cardinal_pythonlib/sql/sql_grammar.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Factory to return an SQL grammer parser, given the name of an SQL dialect.**

"""

from cardinal_pythonlib.sql.sql_grammar import SqlGrammar
from cardinal_pythonlib.sql.sql_grammar_mssql import SqlGrammarMSSQLServer
from cardinal_pythonlib.sql.sql_grammar_mysql import SqlGrammarMySQL
from cardinal_pythonlib.sqlalchemy.dialect import SqlaDialectName


VALID_DIALECTS = [
    SqlaDialectName.MSSQL,
    SqlaDialectName.MYSQL,
    # PostgreSQL not properly supported yet.
]


# =============================================================================
# Factory
# =============================================================================

mysql_grammar = SqlGrammarMySQL()
mssql_grammar = SqlGrammarMSSQLServer()


def make_grammar(dialect: str) -> SqlGrammar:
    """
    Factory to make an :class:`.SqlGrammar` from the name of an SQL dialect,
    where the name is one of the members of :class:`.SqlaDialectName`.
    """
    if dialect == SqlaDialectName.MYSQL:
        return mysql_grammar
    elif dialect == SqlaDialectName.MSSQL:
        return mssql_grammar
    else:
        raise AssertionError(f"Invalid SQL dialect: {dialect!r}")



#!/usr/bin/env python
# cardinal_pythonlib/cache_mw.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**WSGI middleware to disable client-side caching.**

"""

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.wsgi.constants import (
    TYPE_WSGI_APP,
    TYPE_WSGI_APP_RESULT,
    TYPE_WSGI_ENVIRON,
    TYPE_WSGI_EXC_INFO,
    TYPE_WSGI_RESPONSE_HEADERS,
    TYPE_WSGI_START_RESPONSE,
    TYPE_WSGI_START_RESP_RESULT,
    TYPE_WSGI_STATUS,
)

log = get_brace_style_log_with_null_handler(__name__)
# log.setLevel(logging.DEBUG)


# =============================================================================
# DisableClientSideCachingMiddleware
# =============================================================================
# https://stackoverflow.com/questions/49547/making-sure-a-web-page-is-not-cached-across-all-browsers  # noqa
# https://stackoverflow.com/questions/3859097/how-to-add-http-headers-in-wsgi-middleware  # noqa

def add_never_cache_headers(headers: TYPE_WSGI_RESPONSE_HEADERS) -> None:
    """
    Adds WSGI headers to say "never cache this response".
    """
    headers.append(("Cache-Control", "no-cache, no-store, must-revalidate"))  # HTTP 1.1  # noqa
    headers.append(("Pragma", "no-cache"))  # HTTP 1.0
    headers.append(("Expires", "0"))  # Proxies


class DisableClientSideCachingMiddleware(object):
    """
    WSGI middleware to disable client-side caching.
    """

    def __init__(self, app: TYPE_WSGI_APP) -> None:
        self.app = app

    def __call__(self,
                 environ: TYPE_WSGI_ENVIRON,
                 start_response: TYPE_WSGI_START_RESPONSE) \
            -> TYPE_WSGI_APP_RESULT:

        def custom_start_response(status: TYPE_WSGI_STATUS,
                                  headers: TYPE_WSGI_RESPONSE_HEADERS,
                                  exc_info: TYPE_WSGI_EXC_INFO = None) \
                -> TYPE_WSGI_START_RESP_RESULT:
            add_never_cache_headers(headers)
            log.debug("HTTP status {}, headers {}", status, headers)
            return start_response(status, headers, exc_info)

        return self.app(environ, custom_start_response)



#!/usr/bin/env python3
# cardinal_pythonlib/tools/remove_duplicate_files.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Command-line tool to remove duplicate files from a path.**

Largely based on
https://code.activestate.com/recipes/362459-dupinator-detect-and-delete-duplicate-files/  # noqa

"""

from argparse import ArgumentParser
from hashlib import md5
import logging
import os
from pprint import pformat
import stat
from time import sleep
from typing import Dict, List, Union

from cardinal_pythonlib.fileops import gen_filenames
from cardinal_pythonlib.logs import (
    BraceStyleAdapter,
    main_only_quicksetup_rootlogger,
)

log = BraceStyleAdapter(logging.getLogger(__name__))

INITIAL_HASH_SIZE = 1024
MAIN_READ_CHUNK_SIZE = 4096


def deduplicate(directories: List[str], recursive: bool,
                dummy_run: bool) -> None:
    """
    De-duplicate files within one or more directories. Remove files
    that are identical to ones already considered.

    Args:
        directories: list of directories to process
        recursive: process subdirectories (recursively)?
        dummy_run: say what it'll do, but don't do it
    """
    # -------------------------------------------------------------------------
    # Catalogue files by their size
    # -------------------------------------------------------------------------
    files_by_size = {}  # type: Dict[int, List[str]]  # maps size to list of filenames  # noqa
    num_considered = 0
    for filename in gen_filenames(directories, recursive=recursive):
        if not os.path.isfile(filename):
            continue
        size = os.stat(filename)[stat.ST_SIZE]
        a = files_by_size.setdefault(size, [])
        a.append(filename)
        num_considered += 1

    log.debug("files_by_size =\n{}", pformat(files_by_size))

    # -------------------------------------------------------------------------
    # By size, look for duplicates using a hash of the first part only
    # -------------------------------------------------------------------------
    log.info("Finding potential duplicates...")
    potential_duplicate_sets = []
    potential_count = 0
    sizes = list(files_by_size.keys())
    sizes.sort()
    for k in sizes:
        files_of_this_size = files_by_size[k]
        out_files = []  # type: List[str]
        # ... list of all files having >1 file per hash, for this size
        hashes = {}  # type: Dict[bytes, Union[bool, str]]
        # ... key is a hash; value is either True or a filename
        if len(files_of_this_size) == 1:
            continue
        log.info("Testing {} files of size {}...", len(files_of_this_size), k)
        for filename in files_of_this_size:
            if not os.path.isfile(filename):
                continue
            log.debug("Quick-scanning file: {}", filename)
            with open(filename, 'rb') as fd:
                hasher = md5()
                hasher.update(fd.read(INITIAL_HASH_SIZE))
                hash_value = hasher.digest()
                if hash_value in hashes:
                    # We have discovered the SECOND OR SUBSEQUENT hash match.
                    first_file_or_true = hashes[hash_value]
                    if first_file_or_true is not True:
                        # We have discovered the SECOND file;
                        # first_file_or_true contains the name of the FIRST.
                        out_files.append(first_file_or_true)
                        hashes[hash_value] = True
                    out_files.append(filename)
                else:
                    # We have discovered the FIRST file with this hash.
                    hashes[hash_value] = filename
        if out_files:
            potential_duplicate_sets.append(out_files)
            potential_count = potential_count + len(out_files)

    del files_by_size

    log.info("Found {} sets of potential duplicates, based on hashing the "
             "first {} bytes of each...", potential_count, INITIAL_HASH_SIZE)

    log.debug("potential_duplicate_sets =\n{}",
              pformat(potential_duplicate_sets))

    # -------------------------------------------------------------------------
    # Within each set, check for duplicates using a hash of the entire file
    # -------------------------------------------------------------------------
    log.info("Scanning for real duplicates...")

    num_scanned = 0
    num_to_scan = sum(len(one_set) for one_set in potential_duplicate_sets)
    duplicate_sets = []  # type: List[List[str]]
    for one_set in potential_duplicate_sets:
        out_files = []  # type: List[str]
        hashes = {}
        for filename in one_set:
            num_scanned += 1
            log.info("Scanning file [{}/{}]: {}",
                     num_scanned, num_to_scan, filename)
            with open(filename, 'rb') as fd:
                hasher = md5()
                while True:
                    r = fd.read(MAIN_READ_CHUNK_SIZE)
                    if len(r) == 0:
                        break
                    hasher.update(r)
            hash_value = hasher.digest()
            if hash_value in hashes:
                if not out_files:
                    out_files.append(hashes[hash_value])
                out_files.append(filename)
            else:
                hashes[hash_value] = filename
        if len(out_files):
            duplicate_sets.append(out_files)

    log.debug("duplicate_sets = \n{}", pformat(duplicate_sets))

    num_originals = 0
    num_deleted = 0
    for d in duplicate_sets:
        print(f"Original is: {d[0]}")
        num_originals += 1
        for f in d[1:]:
            if dummy_run:
                print(f"Would delete: {f}")
            else:
                print(f"Deleting: {f}")
                os.remove(f)
            num_deleted += 1
        print()

    num_unique = num_considered - (num_originals + num_deleted)
    print(
        "{action} {d} duplicates, leaving {o} originals (and {u} unique files "
        "not touched; {c} files considered in total)".format(
            action="Would delete" if dummy_run else "Deleted",
            d=num_deleted,
            o=num_originals,
            u=num_unique,
            c=num_considered
        )
    )


def main() -> None:
    """
    Command-line processor. See ``--help`` for details.
    """
    parser = ArgumentParser(
        description="Remove duplicate files"
    )
    parser.add_argument(
        "directory", nargs="+",
        help="Files and/or directories to check and remove duplicates from."
    )
    parser.add_argument(
        "--recursive", action="store_true",
        help="Recurse through any directories found"
    )
    parser.add_argument(
        "--dummy_run", action="store_true",
        help="Dummy run only; don't actually delete anything"
    )
    parser.add_argument(
        "--run_repeatedly", type=int,
        help="Run the tool repeatedly with a pause of <run_repeatedly> "
             "seconds between runs. (For this to work well,"
             "you should specify one or more DIRECTORIES in "
             "the 'filename' arguments, not files, and you will need the "
             "--recursive option.)"
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Verbose output"
    )
    args = parser.parse_args()
    main_only_quicksetup_rootlogger(
        level=logging.DEBUG if args.verbose else logging.INFO)

    while True:
        deduplicate(args.directory,
                    recursive=args.recursive,
                    dummy_run=args.dummy_run)
        if args.run_repeatedly is None:
            break
        log.info("Sleeping for {} s...", args.run_repeatedly)
        sleep(args.run_repeatedly)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# cardinal_pythonlib/maths_numpy.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Miscellaneous mathematical functions that use Numpy** (which can be slow to
load).

"""

# =============================================================================
# Imports
# =============================================================================

from collections import Counter
import random
import sys
from typing import List, Optional, Union

import numpy as np  # pip install numpy

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Constants
# =============================================================================

# sys.float_info.max_10_exp:
#       largest integer x such that 10 ** x is representable as a float
# sys.float_info.max_exp:
#       largest integer x such that float(sys.float_info.radix) ** x is
#       representable as a float... and that's 2.0 ** x
# But what we want is the largest integer x such that e ** x = math.exp(x)
# is representable as a float, and that is:

MAX_E_EXPONENT = int(np.log(sys.float_info.max))  # typically, 709


# =============================================================================
# Softmax
# =============================================================================

def softmax(x: np.ndarray,
            b: float = 1.0) -> np.ndarray:
    r"""
    Standard softmax function:

    .. math::

        P_i = \frac {e ^ {\beta \cdot x_i}} { \sum_{i}{\beta \cdot x_i} }

    Args:
        x: vector (``numpy.array``) of values
        b: exploration parameter :math:`\beta`, or inverse temperature
            [Daw2009], or :math:`1/t`; see below

    Returns:
        vector of probabilities corresponding to the input values

    where:

    - :math:`t` is temperature (towards infinity: all actions equally likely;
      towards zero: probability of action with highest value tends to 1)
    - Temperature is not used directly as optimizers may take it to zero,
      giving an infinity; use inverse temperature instead.
    - [Daw2009] Daw ND, "Trial-by-trial data analysis using computational
      methods", 2009/2011; in "Decision Making, Affect, and Learning: Attention
      and Performance XXIII"; Delgado MR, Phelps EA, Robbins TW (eds),
      Oxford University Press.

    """
    n = len(x)
    if b == 0.0:
        # e^0 / sum(a bunch of e^0) = 1 / n
        return np.full(n, 1 / n)

    constant = np.mean(x)
    products = x * b - constant
    # ... softmax is invariant to addition of a constant: Daw article and
    # http://www.faqs.org/faqs/ai-faq/neural-nets/part2/section-12.html#b
    # noinspection PyUnresolvedReferences

    if products.max() > MAX_E_EXPONENT:
        log.warning(
            f"OVERFLOW in softmax(): x = {x}, b = {b}, constant = {constant}, "
            f"x*b - constant = {products}")
        # map the maximum to 1, other things to zero
        index_of_max = np.argmax(products)
        answer = np.zeros(n)
        answer[index_of_max] = 1.0
        return answer

    # noinspection PyUnresolvedReferences
    exponented = np.exp(products)
    sum_exponented = np.sum(exponented)
    if sum_exponented == 0.0:
        # ... avoid division by zero
        return np.full(n, 1 / n)

    return exponented / np.sum(exponented)


def pick_from_probabilities(probabilities: Union[List[float],
                                                 np.ndarray]) -> int:
    """
    Given a list of probabilities like ``[0.1, 0.3, 0.6]``, returns the index
    of the probabilistically chosen item. In this example, we would return
    ``0`` with probability 0.1; ``1`` with probability 0.3; and ``2`` with
    probability 0.6.

    Args:
        probabilities: list of probabilities, which should sum to 1

    Returns:
        the index of the chosen option

    Raises:
        :exc:`ValueError` if a random number in the range [0, 1) is greater
        than or equal to the cumulative sum of the supplied probabilities (i.e.
        if you've specified probabilities adding up to less than 1)

    Does not object if you supply e.g. ``[1, 1, 1]``; it'll always pick the
    first in this example.

    """
    r = random.random()  # range [0.0, 1.0), i.e. 0 <= r < 1
    cs = np.cumsum(probabilities)  # e.g. [0.1, 0.4, 1] in this example
    for idx in range(len(cs)):
        if r < cs[idx]:
            return idx
    raise ValueError(
        f"Probabilities sum to <1: probabilities = {probabilities!r}, "
        f"cumulative sum = {cs!r}"
    )


# =============================================================================
# Logistic
# =============================================================================

def logistic(x: Union[float, np.ndarray],
             k: float,
             theta: float) -> Optional[float]:
    r"""
    Standard logistic function.

    .. math::

        y = \frac {1} {1 + e^{-k (x - \theta)}}

    Args:
        x: :math:`x`
        k: :math:`k`
        theta: :math:`\theta`

    Returns:
        :math:`y`

    """
    # https://www.sharelatex.com/learn/List_of_Greek_letters_and_math_symbols
    if x is None or k is None or theta is None:
        return None
    # noinspection PyUnresolvedReferences
    return 1 / (1 + np.exp(-k * (x - theta)))


def inv_logistic(y: Union[float, np.ndarray],
                 k: float,
                 theta: float) -> Optional[float]:
    r"""
    Inverse standard logistic function:

    .. math::

        x = ( log( \frac {1} {y} - 1) / -k ) + \theta

    Args:
        y: :math:`y`
        k: :math:`k`
        theta: :math:`\theta`

    Returns:
        :math:`x`

    """
    if y is None or k is None or theta is None:
        return None
    # noinspection PyUnresolvedReferences
    return (np.log((1 / y) - 1) / -k) + theta


# =============================================================================
# Testing
# =============================================================================

def _test_softmax() -> None:
    """
    Tests the :func`softmax` function.
    """
    arrays = [
        [1, 1],
        [1, 1, 2],
        [1, 1, 1, 1, 1.01],
        [1, 2, 3, 4, 5],
        [1, 2, 3, 4, 1000],
        [1, 2, 3, 4, 5.0 ** 400],
    ]
    betas = [
        0, 0.5, 1, 2, 10, 100,
    ]
    for x in arrays:
        for b in betas:
            y = softmax(np.array(x), b=b)
            print(f"softmax({x!r}, b={b}) -> {y}")


def _test_pick_from_probabilities() -> None:
    """
    Tests the :func:`pick_from_probabilities` function.
    """
    probabilities = [
        [0.0, 1.0],
        [0.1, 0.3, 0.6],
        [0.25, 0.25, 5],
        # [0.25],  # crashes (appropriately)
        [1, 1, 1],  # silly, but doesn't crash
    ]
    n_values = [10, 1000, 100000]
    for p in probabilities:
        for n in n_values:
            c = Counter()
            c.update(pick_from_probabilities(p) for _ in range(n))
            sc = sorted(c.items(), key=lambda kv: kv[0])
            print(f"_test_pick_from_probabilities: p = {p}, n = {n} -> {sc}")


if __name__ == '__main__':
    _test_softmax()
    _test_pick_from_probabilities()



#!/usr/bin/env python
# cardinal_pythonlib/signalfunc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions to handle OS signals that may cause trouble.**

"""

import platform
import signal

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Signal handlers
# =============================================================================

# noinspection PyUnusedLocal
def ctrl_c_trapper(signum: int, stackframe) -> None:
    """
    Logs that ``CTRL-C`` has been pressed but does nothing else.
    """
    log.critical("Ignoring CTRL+C (signal {}); use the GUI to quit", signum)


# noinspection PyUnusedLocal
def ctrl_break_trapper(signum: int, stackframe) -> None:
    """
    Logs that ``CTRL-BREAK`` has been pressed but does nothing else.
    """
    log.critical("Ignoring CTRL+BREAK (signal {}); use the GUI to quit",
                 signum)


# noinspection PyUnusedLocal
def sigterm_trapper(signum: int, stackframe) -> None:
    """
    Logs that ``SIGTERM`` has been received but does nothing else.
    """
    log.critical("Ignoring SIGTERM (signal {}); use the GUI to quit", signum)


def trap_ctrl_c_ctrl_break() -> None:
    """
    Prevent ``CTRL-C``, ``CTRL-BREAK``, and similar signals from doing
    anything.

    See

    - https://docs.python.org/3/library/signal.html#signal.SIG_IGN
    - https://msdn.microsoft.com/en-us/library/xdkz3x12.aspx
    - https://msdn.microsoft.com/en-us/library/windows/desktop/ms682541(v=vs.85).aspx

    Under Windows, the only options are:

      =========== ======================= =====================================
      Signal      Meaning                 Comment
      =========== ======================= =====================================
      SIGABRT     abnormal termination
      SIGFPE      floating-point error
      SIGILL      illegal instruction
      SIGINT      CTRL+C signal           -- trapped here
      SIGSEGV     illegal storage access
      SIGTERM     termination request     -- trapped here
      SIGBREAK    CTRL+BREAK              -- trapped here under Windows
      =========== ======================= =====================================

    In Linux, you also find:

      =========== =============================
      Signal      Meaning
      =========== =============================
      SIGBUS      bus error / unaligned access
      =========== =============================

    To ignore, can do:

    .. code-block:: python

      signal.signal(signal.SIGINT, signal.SIG_IGN)  # SIG_IGN = "ignore me"

    or pass a specified handler, as in the code here.
    """  # noqa

    signal.signal(signal.SIGINT, ctrl_c_trapper)
    signal.signal(signal.SIGTERM, sigterm_trapper)
    if platform.system() == 'Windows':
        # SIGBREAK isn't in the Linux signal module
        # noinspection PyUnresolvedReferences
        signal.signal(signal.SIGBREAK, ctrl_break_trapper)



#!/usr/bin/env python
# cardinal_pythonlib/classes.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to help work with Python classes.**

"""

from typing import Any, Dict, Generator, List, Type, TypeVar


# =============================================================================
# Does a derived class implement a method?
# =============================================================================

"""
https://stackoverflow.com/questions/1776994
https://docs.python.org/3/library/inspect.html
https://github.com/edoburu/django-fluent-contents/issues/43
https://bytes.com/topic/python/answers/843424-python-2-6-3-0-determining-if-method-inherited  # noqa
https://docs.python.org/3/reference/datamodel.html

In Python 2, you can do this:
    return derived_method.__func__ != base_method.__func__
In Python 3.4:
    ...

class Base(object):
    def one():
        print("base one")
    def two():
        print("base two")


class Derived(Base):
    def two():
        print("derived two")


Derived.two.__dir__()  # not all versions of Python


derived_class_implements_method(Derived, Base, 'one')  # should be False
derived_class_implements_method(Derived, Base, 'two')  # should be True
derived_class_implements_method(Derived, Base, 'three')  # should be False

"""

T1 = TypeVar('T1')
T2 = TypeVar('T2')


def derived_class_implements_method(derived: Type[T1],
                                    base: Type[T2],
                                    method_name: str) -> bool:
    """
    Does a derived class implement a method (and not just inherit a base
    class's version)?

    Args:
        derived: a derived class
        base: a base class
        method_name: the name of a method

    Returns:
        whether the derived class method is (a) present, and (b) different to
        the base class's version of the same method

    Note: if C derives from B derives from A, then a check on C versus A will
    succeed if C implements the method, or if C inherits it from B but B has
    re-implemented it compared to A.

    """
    derived_method = getattr(derived, method_name, None)
    if derived_method is None:
        return False
    base_method = getattr(base, method_name, None)
    # if six.PY2:
    #     return derived_method.__func__ != base_method.__func__
    # else:
    #     return derived_method is not base_method
    return derived_method is not base_method


# =============================================================================
# Subclasses
# =============================================================================
# https://stackoverflow.com/questions/3862310/how-can-i-find-all-subclasses-of-a-class-given-its-name  # noqa

def gen_all_subclasses(cls: Type) -> Generator[Type, None, None]:
    """
    Generates all subclasses of a class.

    Args:
        cls: a class

    Yields:
        all subclasses

    """

    for s1 in cls.__subclasses__():
        yield s1
        for s2 in gen_all_subclasses(s1):
            yield s2


def all_subclasses(cls: Type) -> List[Type]:
    """
    Returns all subclasses of a class.

    Args:
        cls: a class

    Returns:
        a list of all subclasses

    """
    return list(gen_all_subclasses(cls))


# =============================================================================
# Class properties
# =============================================================================

class ClassProperty(property):
    """
    One way to mark a function as a class property (logically, a combination of
    ``@classmethod`` and ``@property``).

    See
    https://stackoverflow.com/questions/128573/using-property-on-classmethods.

    However, in practice we use :class:`classproperty`, a slightly different
    version.
    """
    # https://stackoverflow.com/questions/128573/using-property-on-classmethods
    # noinspection PyMethodOverriding
    def __get__(self, cls, owner):
        # noinspection PyUnresolvedReferences
        return self.fget.__get__(None, owner)()


# noinspection PyPep8Naming
class classproperty(object):
    """
    Decorator to mark a function as a class property (logically, a combination
    of ``@classmethod`` and ``@property``).

    See
    https://stackoverflow.com/questions/128573/using-property-on-classmethods
    """
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


# =============================================================================
# Class attributes and their values
# =============================================================================

def class_attribute_dict(
        cls: Type,
        exclude_underscore: bool = True,
        exclude_double_underscore: bool = True) -> Dict[str, Any]:
    """
    When given a class, returns a dictionary of all its attributes, by default
    excluding those starting with single and double underscores.

    This is just a filtered versio of cls.__dict__.
    """
    d = {}  # type: Dict[str, Any]
    for k, v in cls.__dict__.items():
        if exclude_underscore and k.startswith("_"):
            continue
        if exclude_double_underscore and k.startswith("__"):
            continue
        d[k] = v
    return d


def class_attribute_names(cls: Type,
                          exclude_underscore: bool = True,
                          exclude_double_underscore: bool = True) -> List[str]:
    """
    When given a class, returns the NAMES of all its attributes, by default
    excluding those starting with single and double underscores.

    Used in particular to enumerate constants provided within a class.
    """
    d = class_attribute_dict(
        cls,
        exclude_underscore=exclude_underscore,
        exclude_double_underscore=exclude_double_underscore
    )
    return list(d.keys())


def class_attribute_values(cls: Type,
                           exclude_underscore: bool = True,
                           exclude_double_underscore: bool = True) -> List[str]:
    """
    When given a class, returns the VALUES of all its attributes, by default
    excluding those starting with single and double underscores.

    Used in particular to enumerate constants provided within a class.
    """
    d = class_attribute_dict(
        cls,
        exclude_underscore=exclude_underscore,
        exclude_double_underscore=exclude_double_underscore
    )
    return list(d.values())



#!/usr/bin/env python
# cardinal_pythonlib/configfiles.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for config (.INI) file reading.**

"""

from configparser import ConfigParser, NoOptionError
import logging
from typing import Any, Callable, Iterable, List

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Style 1
# =============================================================================

def get_config_string_option(parser: ConfigParser,
                             section: str,
                             option: str,
                             default: str = None) -> str:
    """
    Retrieves a string value from a parser.

    Args:
        parser: instance of :class:`ConfigParser`
        section: section name within config file
        option: option (variable) name within that section
        default: value to return if option is absent

    Returns:
        string value

    Raises:
        ValueError: if the section is absent

    """
    if not parser.has_section(section):
        raise ValueError("config missing section: " + section)
    return parser.get(section, option, fallback=default)


def read_config_string_options(obj: Any,
                               parser: ConfigParser,
                               section: str,
                               options: Iterable[str],
                               default: str = None) -> None:
    """
    Reads config options and writes them as attributes of ``obj``, with
    attribute names as per ``options``.

    Args:
        obj: the object to modify
        parser: instance of :class:`ConfigParser`
        section: section name within config file
        options: option (variable) names within that section
        default: value to use for any missing options

    Returns:

    """
    # enforce_str removed; ConfigParser always returns strings unless asked
    # specifically
    for o in options:
        setattr(obj, o, get_config_string_option(parser, section, o,
                                                 default=default))


def get_config_multiline_option(parser: ConfigParser,
                                section: str,
                                option: str,
                                default: List[str] = None) -> List[str]:
    """
    Retrieves a multi-line string value from a parser as a list of strings
    (one per line, ignoring blank lines).

    Args:
        parser: instance of :class:`ConfigParser`
        section: section name within config file
        option: option (variable) name within that section
        default: value to return if option is absent (``None`` is mapped to
            ``[]``)

    Returns:
        list of strings

    Raises:
        ValueError: if the section is absent

    """
    default = default or []
    if not parser.has_section(section):
        raise ValueError("config missing section: " + section)
    try:
        multiline = parser.get(section, option)
        values = [x.strip() for x in multiline.splitlines() if x.strip()]
        return values
    except NoOptionError:
        return default


def read_config_multiline_options(obj: Any,
                                  parser: ConfigParser,
                                  section: str,
                                  options: Iterable[str]) -> None:
    """
    This is to :func:`read_config_string_options` as
    :func:`get_config_multiline_option` is to :func:`get_config_string_option`.
    """
    for o in options:
        setattr(obj, o, get_config_multiline_option(parser, section, o))


def get_config_bool_option(parser: ConfigParser,
                           section: str,
                           option: str,
                           default: bool = None) -> bool:
    """
    Retrieves a boolean value from a parser.

    Args:
        parser: instance of :class:`ConfigParser`
        section: section name within config file
        option: option (variable) name within that section
        default: value to return if option is absent

    Returns:
        Boolean value

    Raises:
        ValueError: if the section is absent

    """
    if not parser.has_section(section):
        raise ValueError("config missing section: " + section)
    return parser.getboolean(section, option, fallback=default)


# =============================================================================
# Style 2
# =============================================================================

# =============================================================================
# Reading config files: style 2
# =============================================================================

def get_config_parameter(config: ConfigParser,
                         section: str,
                         param: str,
                         fn: Callable[[Any], Any],
                         default: Any = None,
                         loglevel: int = logging.DEBUG) -> Any:
    """
    Fetch parameter from ``configparser`` ``.INI`` file.

    Args:
        config: :class:`ConfigParser` object
        section: section name within config file
        param: name of parameter within section
        fn: function to apply to string parameter (e.g. ``int``)
        default: default value
        loglevel: log level if default is needed

    Returns:
        parameter value, or ``None`` if ``default is None``, or ``fn(default)``
    """
    try:
        value = fn(config.get(section, param))
    except (TypeError, ValueError, NoOptionError):
        log.log(
            level=loglevel,
            msg=f"Configuration variable {param} not found or improper in "
                f"section [{section}]; using default of {default!r}")
        if default is None:
            value = default
        else:
            value = fn(default)
    return value


def get_config_parameter_boolean(config: ConfigParser,
                                 section: str,
                                 param: str,
                                 default: bool,
                                 loglevel: int = logging.DEBUG) -> bool:
    """
    Get Boolean parameter from ``configparser`` ``.INI`` file.

    Args:
        config: :class:`ConfigParser` object
        section: section name within config file
        param: name of parameter within section
        default: default value
        loglevel: log level if default is needed
    Returns:
        parameter value, or default
    """
    try:
        value = config.getboolean(section, param)
    except (TypeError, ValueError, NoOptionError):
        log.log(
            level=loglevel,
            msg=f"Configuration variable {param} not found or improper in "
                f"section [{section}]; using default of {default!r}")
        value = default
    return value


def get_config_parameter_loglevel(config: ConfigParser,
                                  section: str,
                                  param: str,
                                  default: int,
                                  loglevel: int = logging.DEBUG) -> int:
    """
    Get ``loglevel`` parameter from ``configparser`` ``.INI`` file, e.g.
    mapping ``'debug'`` to ``logging.DEBUG``.

    Args:
        config: :class:`ConfigParser` object
        section: section name within config file
        param: name of parameter within section
        default: default value
        loglevel: log level if default is needed
    Returns:
        parameter value, or default
    """
    try:
        value = config.get(section, param).lower()
        if value == "debug":
            return logging.DEBUG  # 10
        elif value == "info":
            return logging.INFO
        elif value in ["warn", "warning"]:
            return logging.WARN
        elif value == "error":
            return logging.ERROR
        elif value in ["critical", "fatal"]:
            return logging.CRITICAL  # 50
        else:
            raise ValueError
    except (TypeError, ValueError, NoOptionError, AttributeError):
        log.log(
            level=loglevel,
            msg=f"Configuration variable {param} not found or improper in "
                f"section [{section}]; using default of {default!r}")
        return default


def get_config_parameter_multiline(config: ConfigParser,
                                   section: str,
                                   param: str,
                                   default: List[str],
                                   loglevel: int = logging.DEBUG) -> List[str]:
    """
    Get multi-line string parameter from ``configparser`` ``.INI`` file,
    as a list of strings (one per line, ignoring blank lines).

    Args:
        config: :class:`ConfigParser` object
        section: section name within config file
        param: name of parameter within section
        default: default value
        loglevel: log level if default is needed
    Returns:
        parameter value, or default
    """
    try:
        multiline = config.get(section, param)
        lines = [x.strip() for x in multiline.splitlines()]
        return [line for line in lines if line]
    except (TypeError, ValueError, NoOptionError):
        log.log(
            level=loglevel,
            msg=f"Configuration variable {param} not found or improper in "
                f"section [{section}]; using default of {default!r}")
        return default
